import statistics as st
import os
import numpy as np
import copy
import matplotlib.pyplot as plt
from matplotlib import cm
from tifffile.tifffile import imwrite
import matplotlib.patches as mpatches
from PIL import Image
from NaroNet.utils.parallel_process import parallel_process
from NaroNet.BioInsights.Pheno_Neigh_Info import load_cell_types_assignments, load_patch_image
import gc
import cv2

def TME_location_in_image_(dataset,thisfolder,idxclster,patchIDX, clusters,nClust,sts,statisticalTests_PerPatient,statisticalTests,unrestrictedLoss,count,Patindx):
    '''
    '''

    if not os.path.exists(dataset.bioInsights_dir_TME_in_image+thisfolder+idxclster[2][0]+'/'):
        os.mkdir(dataset.bioInsights_dir_TME_in_image+thisfolder+idxclster[2][0]+'/')

    # Pixel-to-cluster Matrix
    clust0 = np.load(dataset.processed_dir_cell_types+'cluster_assignmentPerPatch_Index_{}_{}_ClustLvl_{}.npy'.format(idxclster[1], 0, clusters[0]))                    
    if nClust>0:
        clust1 = np.load(dataset.processed_dir_cell_types+'cluster_assignmentPerPatch_Index_{}_{}_ClustLvl_{}.npy'.format(idxclster[1], 0, clusters[1]))                                     
    if nClust>1:
        clust2 = np.load(dataset.processed_dir_cell_types+'cluster_assignment_Index_{}_ClustLvl_{}.npy'.format(idxclster[1], clusters[2]))                    
        clust2 = np.matmul(clust1,clust2)
    clust = clust0 
    clust = clust1 if nClust==1 else clust
    clust = clust2 if nClust==2 else clust                
    
    # Open Raw Image.
    im, imList = dataset.open_Raw_Image(idxclster,1)

    # Assign significant cluster as 1 to SuperPatch Image             
    if ('Endometrial_LowGrade' in dataset.raw_dir):
        for imList_n ,imList_i in enumerate(imList[::-1]):
            Patch_im_i = np.zeros(imList_i.shape[:2])                        
            # Create superPatch Label image using the Superpatch size. 
            division_rows = np.floor(Patch_im_i.shape[0]/dataset.patch_size)
            division_cols = np.floor(Patch_im_i.shape[1]/dataset.patch_size)
            lins = np.repeat(list(range(int(division_cols))), dataset.patch_size)
            lins = np.repeat(np.expand_dims(lins,axis=1), dataset.patch_size, axis=1)
            for row_indx, row in enumerate(range(int(division_rows))):
                Patch_im_i[row_indx*dataset.patch_size:(row_indx+1)*dataset.patch_size,:int(division_cols*dataset.patch_size)] = np.transpose(lins+int(row_indx*division_cols))
            Patch_im_i = Patch_im_i.astype(int) 
            if imList_n==0:
                Patch_im = Patch_im_i
            else:                
                Patch_im = np.concatenate((Patch_im, Patch_im_i+Patch_im.max()+1),1)
    elif 'ZuriBasel' in dataset.raw_dir:
        for imList_n ,imList_i in enumerate(imList[::-1]):
            Patch_im_i = np.zeros(imList_i.shape[:2])                        
            # Create superPatch Label image using the Superpatch size. 
            division_rows = np.floor(Patch_im_i.shape[0]/dataset.patch_size)
            division_cols = np.floor(Patch_im_i.shape[1]/dataset.patch_size)
            lins = np.repeat(list(range(int(division_cols))), dataset.patch_size)
            lins = np.repeat(np.expand_dims(lins,axis=1), dataset.patch_size, axis=1)
            for row_indx, row in enumerate(range(int(division_rows))):
                Patch_im_i[row_indx*dataset.patch_size:(row_indx+1)*dataset.patch_size,:int(division_cols*dataset.patch_size)] = np.transpose(lins+int(row_indx*division_cols))
            Patch_im_i = Patch_im_i.astype(int) 
            if imList_n==0:
                Patch_im = Patch_im_i
            else:                
                Patch_im = np.concatenate((Patch_im, Patch_im_i+Patch_im.max()+1),1)
    else:
        Patch_im = np.zeros(im.shape[:2])                        
        # Create superPatch Label image using the Superpatch size. 
        division_rows = np.floor(Patch_im.shape[0]/dataset.patch_size)
        division_cols = np.floor(Patch_im.shape[1]/dataset.patch_size)
        lins = np.repeat(list(range(int(division_cols))), dataset.patch_size)
        lins = np.repeat(np.expand_dims(lins,axis=1), dataset.patch_size, axis=1)
        for row_indx, row in enumerate(range(int(division_rows))):
            Patch_im[row_indx*dataset.patch_size:(row_indx+1)*dataset.patch_size,:int(division_cols*dataset.patch_size)] = np.transpose(lins+int(row_indx*division_cols))
        # suprpxlIm = np.transpose(suprpxlIm)
        Patch_im = Patch_im.astype(int)
        if ('V_H' in dataset.root) or ('V3' in dataset.root) or ('V1' in dataset.root) or ('V4' in dataset.root):
            Patch_im = np.transpose(Patch_im)                        

    # Assign significant clusters to pixels.
    cell_type_top1 = np.apply_along_axis(np.argmax, axis=1, arr=clust)       

    # Obtain Image in RGB
    # imRGB = dataset.nPlex2RGB(im)                                   
    # for c in range(imRGB.shape[2]):
    #     imRGB[:,:,c] = imRGB[:,:,c]/imRGB[:,:,c].max()                 
                            
    # For each threshold value
    #for thrs in ClusteringThreshold:
    # Mask selecting top PIR values.
    AllClusters = (copy.deepcopy(cell_type_top1)*0).astype('float32')
    AllClusters[sts[2][1]==cell_type_top1] = clust[sts[2][1]==cell_type_top1,sts[2][1]]
    # Apply Threshold to select cells with high confidence.
    # AllClusters[np.percentile(clust.max(-1),thrs)>clust.max(-1)] = 0 
    AllClusters_2=AllClusters[Patch_im] # Superpatches of significant clusters.
    # Number of SuperPatches present in this one    
    # Save Certainty of this Phenotype-TissueCommunity     
    plt.figure()
    n, bins, patches = plt.hist(clust.max(-1)[sts[2][1]==cell_type_top1], 100, color=cm.jet_r(int(sts[2][1]*(255/int(sts[2][0])))), alpha=0.5)

    if int(sts[2][0])==clusters[0]:
        plt.xlabel('Level of Phenotype Certainty')
        plt.title('Histogram of Phenotype '+str(sts[2][1])+' Certainty')
    elif int(sts[2][0])==clusters[1]:
        plt.xlabel('Level of Neighborhood Certainty')
        plt.title('Histogram of Neighborhood '+str(sts[2][1])+' Certainty')
    elif int(sts[2][0])==clusters[2]:
        plt.xlabel('Level of Area Certainty')
        plt.title('Histogram of Area '+str(sts[1])+' Certainty')
    plt.ylabel('Number of patches')
    plt.savefig(dataset.bioInsights_dir_TME_in_image+thisfolder+idxclster[2][0]+'/Label{}_Slide{}_Patch{}_Clstrs{}_Thrs{}_Acc{}_PIR{}_Hist.png'.format(
        idxclster[2][0],idxclster[0],patchIDX,statisticalTests['TME -h'][count],100,unrestrictedLoss[count],str(round(statisticalTests_PerPatient[Patindx][1],2))), format="PNG",dpi=200) 
    plt.close()
    # Save GT and Image                           
    # imwrite(dataset.bioInsights_dir_TME_in_image+thisfolder+idxclster[2][0]+'/Label{}_Slide{}_Patch{}_Clstrs{}_Acc{}_PIR{}_Images.tiff'.format(idxclster[2][0],idxclster[0],patchIDX,statisticalTests['TME -h'][count],unrestrictedLoss[count],str(round(statisticalTests_PerPatient[Patindx][1],2))),np.moveaxis(im,2,0))                                        
    imtosave = Image.fromarray(np.uint8(AllClusters_2*255))
    imtosave.save(dataset.bioInsights_dir_TME_in_image+thisfolder+idxclster[2][0]+'/Label{}_Slide{}_Patch{}_Clstrs{}_Acc{}_PIR{}_Label.tiff'.format(idxclster[2][0],idxclster[0],patchIDX,statisticalTests['TME -h'][count],unrestrictedLoss[count],str(round(statisticalTests_PerPatient[Patindx][1],2))))                
    
    return 'done'

def TME_location_in_image(dataset, statisticalTests, clusters, IndexAndClass,unrestrictedLoss,statisticalTests_PerPatient, num_classes, attentionLayer,ClusteringThreshold):
    '''
        docstring
    '''
    
    # statisticalTests = sorted(statisticalTests, key=lambda k: k[0]) # 1.p-value, 2.Cluster step, 3. column of the heatmap        
    # stsTest=statisticalTests[0]
    IntersecIndex=[]
    patchIDX = dataset.args['epochs']
    dict_subjects = []
    for nClust, clusterToProcess in enumerate(clusters):            
        IntersecIndex.append([])
        # For each image...
        for count, Patindx in enumerate(statisticalTests['Patient index']):      
            idxclster = IndexAndClass[Patindx]                   
            thisfolder = statisticalTests['TME -h'][count]+'/'
            if not os.path.exists(dataset.bioInsights_dir_TME_in_image+thisfolder):
                os.mkdir(dataset.bioInsights_dir_TME_in_image+thisfolder)            
            # One Phenotype or Tissue Community to one image
            for sts in statisticalTests['TME'][count]:
                if int(sts[2][0])==clusterToProcess: 
                    dict_subjects.append({'dataset':dataset,'thisfolder':thisfolder,'idxclster':idxclster,'patchIDX':patchIDX, 
                                        'clusters':clusters,'nClust':nClust,'sts':sts,'statisticalTests_PerPatient':statisticalTests_PerPatient,
                                        'statisticalTests':statisticalTests,'unrestrictedLoss':unrestrictedLoss,'count':count,'Patindx':Patindx})
    
    # select_patches_from_cohort
    result = parallel_process(dict_subjects,TME_location_in_image_,use_kwargs=True,front_num=1,desc='BioInsights: Save TOP-k PIRs for each TME') 
    return 1

def TissueArea_in_image_(dataset, thisfolder, subject,Area_conf):
    # Open Raw Image.
    im, imList = dataset.open_Raw_Image(subject,1)
    
    # Assign significant cluster as 1 to SuperPatch Image             
    if ('Endometrial_LowGrade' in dataset.raw_dir):
        for imList_n ,imList_i in enumerate(imList[::-1]):
            Patch_im_i = np.zeros(imList_i.shape[:2])                        
            # Create superPatch Label image using the Superpatch size. 
            division_rows = np.floor(Patch_im_i.shape[0]/dataset.patch_size)
            division_cols = np.floor(Patch_im_i.shape[1]/dataset.patch_size)
            lins = np.repeat(list(range(int(division_cols))), dataset.patch_size)
            lins = np.repeat(np.expand_dims(lins,axis=1), dataset.patch_size, axis=1)
            for row_indx, row in enumerate(range(int(division_rows))):
                Patch_im_i[row_indx*dataset.patch_size:(row_indx+1)*dataset.patch_size,:int(division_cols*dataset.patch_size)] = np.transpose(lins+int(row_indx*division_cols))
            Patch_im_i = Patch_im_i.astype(int) 
            if imList_n==0:
                Patch_im = Patch_im_i
            else:                
                Patch_im = np.concatenate((Patch_im, Patch_im_i+Patch_im.max()+1),1)
    elif ('Synthetic' in dataset.raw_dir):
        for imList_n ,imList_i in enumerate(imList[::-1]):
            Patch_im_i = np.zeros(imList_i.shape[:2])                        
            # Create superPatch Label image using the Superpatch size. 
            division_rows = np.floor(Patch_im_i.shape[0]/dataset.patch_size)
            division_cols = np.floor(Patch_im_i.shape[1]/dataset.patch_size)
            lins = np.repeat(list(range(int(division_cols))), dataset.patch_size)
            lins = np.repeat(np.expand_dims(lins,axis=1), dataset.patch_size, axis=1)
            for row_indx, row in enumerate(range(int(division_rows))):
                Patch_im_i[row_indx*dataset.patch_size:(row_indx+1)*dataset.patch_size,:int(division_cols*dataset.patch_size)] = np.transpose(lins+int(row_indx*division_cols))
            Patch_im_i = Patch_im_i.astype(int) 
            if imList_n==0:
                Patch_im = Patch_im_i
            else:                
                Patch_im = np.concatenate((Patch_im, Patch_im_i+Patch_im.max()+1),1)
    elif 'Endometrial_POLE' in dataset.root:
        Patch_im = np.zeros(im.shape[:2])                        
        # Create superPatch Label image using the Superpatch size. 
        division_rows = np.floor(Patch_im.shape[0]/dataset.patch_size)
        division_cols = np.floor(Patch_im.shape[1]/dataset.patch_size)
        lins = np.repeat(list(range(int(division_cols))), dataset.patch_size)
        lins = np.repeat(np.expand_dims(lins,axis=1), dataset.patch_size, axis=1)
        for row_indx, row in enumerate(range(int(division_rows))):
            Patch_im[row_indx*dataset.patch_size:(row_indx+1)*dataset.patch_size,:int(division_cols*dataset.patch_size)] = np.transpose(lins+int(row_indx*division_cols))
        # Patch_im = np.transpose(Patch_im)
        Patch_im = Patch_im.astype(int)
    else: 
        Patch_im = np.zeros(im.shape[:2])                        
        # Create superPatch Label image using the Superpatch size. 
        division_rows = np.floor(Patch_im.shape[0]/dataset.patch_size)
        division_cols = np.floor(Patch_im.shape[1]/dataset.patch_size)
        lins = np.repeat(list(range(int(division_cols))), dataset.patch_size)
        lins = np.repeat(np.expand_dims(lins,axis=1), dataset.patch_size, axis=1)
        for row_indx, row in enumerate(range(int(division_rows))):
            Patch_im[row_indx*dataset.patch_size:(row_indx+1)*dataset.patch_size,:int(division_cols*dataset.patch_size)] = np.transpose(lins+int(row_indx*division_cols))
        # Patch_im = np.transpose(Patch_im)
        Patch_im = Patch_im.astype(int)
    
    # Superpatches of significant clusters.
    Area_conf_im = Area_conf[Patch_im] 
    Area_conf_im = Area_conf_im/Area_conf_im.max()
    # Save tissue area confidence
    imtosave = Image.fromarray(np.uint8(Area_conf_im*255))
    imtosave.save(dataset.bioInsights_dir_TME_in_image+thisfolder+'{}_conf{}_Label.tiff'.format(subject[0],round(Area_conf.max(),3)))                
    # Save image original.
    imwrite(dataset.bioInsights_dir_TME_in_image+thisfolder+'{}_conf{}_Images.tiff'.format(subject[0],round(Area_conf.max(),3)),np.moveaxis(im,2,0))                                        


def TissueArea_in_image(dataset,TissueArea_Conf,subjects,top_k):
    '''
        docstring
    '''   
    dict_subjects = []

    # choose highest confidence tissue areas.
    for n_ar, tissue_area in enumerate(TissueArea_Conf):
        thisfolder = 'A{}/'.format(str(n_ar+1))
        if not os.path.exists(dataset.bioInsights_dir_TME_in_image+thisfolder):
            os.mkdir(dataset.bioInsights_dir_TME_in_image+thisfolder)            

        max_conf_for_area = [t[0].max() for t in tissue_area] 
        hi_conf_subjcts = sorted(range(len(max_conf_for_area)), key=lambda i: max_conf_for_area[i])[:top_k]                    
        for hi_conf_s in hi_conf_subjcts:
            dict_subjects.append({'dataset':dataset, 'subject':subjects[hi_conf_s],'Area_conf':tissue_area[hi_conf_s]})                
                         
    # select_patches_from_cohort
    result = parallel_process(dict_subjects,TissueArea_in_image_,use_kwargs=True,front_num=1,desc='BioInsights: Save examples of tissue areas') 
    return 1

def All_TMEs_in_Image_(clusters,dataset,subject_info):
    
    # Crops,Confidence,and Phenotype Vector.    
    prev_cell_type_assignment=[]
        
    # Apply mask to patch
    for subgraph_idx in range(dataset.findLastIndex(subject_info[1])+1):                
        
        # Open Raw Image.        
        im, imList = dataset.open_Raw_Image(subject_info,1)                

        # Load image of patches
        Patch_im,Patch_IMS = load_patch_image(dataset,imList)                    
        
        # Load Patch Contrastive Learning Information
        PCL_reprsntions = np.load(dataset.raw_dir+'/{}.npy'.format(subject_info[0])) 

        for cell_type_idx, n_cell_types in enumerate(clusters):    
            # load cell_types_assignments
            prev_cell_type_assignment, cell_type_assignment = load_cell_types_assignments(
                                                                    dataset,cell_type_idx,subject_info,subgraph_idx,
                                                                    n_cell_types,prev_cell_type_assignment)                                    
            
            # Initialize labels
            Labels = np.zeros((im.shape[0],im.shape[1],n_cell_types),dtype=np.uint8)

            # For each cell type
            for cell_type in range(n_cell_types):
                
                # Create image with probabilities with the size of the original image
                aux = (cell_type_assignment[:Patch_im.max()+1,cell_type]*255).astype(np.uint8)                
                last_patch = 0
                aux_2 = []
                for im_i in Patch_IMS:
                    aux_2.append(np.reshape(aux[last_patch:last_patch+im_i.max()+1],(im_i.shape[1],im_i.shape[0])))                    
                    last_patch = last_patch+im_i.max()+1
                aux = np.concatenate(aux_2,1)                

                # Interpolate aux image to match size with the original one.
                Labels[:,:,cell_type] = cv2.resize(aux, (im.shape[1],im.shape[0]), 0, 0, interpolation = cv2.INTER_NEAREST)        

            # Save image of labels as a multitiff.
            if cell_type_idx==0:
                imwrite(dataset.bioInsights_dir_TME_in_image+'{}_Phenotypesss.tif'.format(subject_info[0]),np.moveaxis(Labels,2,0),photometric='minisblack')                
            elif cell_type_idx==1:
                imwrite(dataset.bioInsights_dir_TME_in_image+'{}_Neighborhoods.tif'.format(subject_info[0]), np.moveaxis(Labels,2,0), photometric='minisblack')        
            else:
                imwrite(dataset.bioInsights_dir_TME_in_image+'{}_Areas.tif'.format(subject_info[0]), np.moveaxis(Labels,2,0), photometric='minisblack')       

            # Save original image
            if dataset.args['Bio_save_Orig_im']:                
                imwrite(dataset.bioInsights_dir_TME_in_image+'{}.tif'.format(subject_info[0]), np.moveaxis(np.concatenate(imList,1),2,0), photometric='minisblack')  


        del im, imList, PCL_reprsntions, Patch_im
        gc.collect(0)
        gc.collect(1)
        gc.collect(2)


def All_TMEs_in_Image(dataset,clusters, IndexAndClass):
    # Apply mask to slide
    dict_subjects = []
    for count, subject_info in enumerate(IndexAndClass):
        dict_subjects.append({'clusters':clusters,'dataset':dataset,'subject_info':subject_info})
    
    result = parallel_process(dict_subjects,All_TMEs_in_Image_,use_kwargs=True,front_num=5,desc='BioInsights: Save TMEs for each patient') 



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
import torch.nn as nn

class GCN(nn.Module):
    """ Graph convolution unit (single layer)
    """

    def __init__(self, num_state, num_node, bias=False):
        super(GCN, self).__init__()
        self.conv1 = nn.Conv1d(num_node, num_node, kernel_size=1)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv1d(num_state, num_state, kernel_size=1, bias=bias)

    def forward(self, x):
        # (n, num_state, num_node) -> (n, num_node, num_state)
        #                          -> (n, num_state, num_node)
        h = self.conv1(x.permute(0, 2, 1).contiguous()).permute(0, 2, 1)
        h = h + x
        # (n, num_state, num_node) -> (n, num_state, num_node)
        h = self.conv2(self.relu(h))
        return h


class GloRe_Unit(nn.Module):
    """
    Graph-based Global Reasoning Unit
    Parameter:
        'normalize' is not necessary if the input size is fixed
    """
    def __init__(self, num_in, num_mid, 
                 ConvNd=nn.Conv3d,
                 BatchNormNd=nn.BatchNorm3d,
                 normalize=False):
        super(GloRe_Unit, self).__init__()
        
        self.normalize = normalize
        self.num_s = int(2 * num_mid)
        self.num_n = int(1 * num_mid)

        # reduce dim
        self.conv_state = ConvNd(num_in, self.num_s, kernel_size=1)
        # projection map
        self.conv_proj = ConvNd(num_in, self.num_n, kernel_size=1)
        # ----------
        # reasoning via graph convolution
        self.gcn = GCN(num_state=self.num_s, num_node=self.num_n)
        # ----------
        # extend dimension
        self.conv_extend = ConvNd(self.num_s, num_in, kernel_size=1, bias=False)

        self.blocker = BatchNormNd(num_in, eps=1e-04) # should be zero initialized


    def forward(self, x):
        '''
        :param x: (n, c, d, h, w)
        '''
        n = x.size(0)

        # (n, num_in, h, w) --> (n, num_state, h, w)
        #                   --> (n, num_state, h*w)
        x_state_reshaped = self.conv_state(x).view(n, self.num_s, -1)

        # (n, num_in, h, w) --> (n, num_node, h, w)
        #                   --> (n, num_node, h*w)
        x_proj_reshaped = self.conv_proj(x).view(n, self.num_n, -1)

        # (n, num_in, h, w) --> (n, num_node, h, w)
        #                   --> (n, num_node, h*w)
        x_rproj_reshaped = x_proj_reshaped

        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

        # projection: coordinate space -> interaction space
        # (n, num_state, h*w) x (n, num_node, h*w)T --> (n, num_state, num_node)
        x_n_state = torch.matmul(x_state_reshaped, x_proj_reshaped.permute(0, 2, 1))
        if self.normalize:
            x_n_state = x_n_state * (1. / x_state_reshaped.size(2))

        # reasoning: (n, num_state, num_node) -> (n, num_state, num_node)
        x_n_rel = self.gcn(x_n_state)

        # reverse projection: interaction space -> coordinate space
        # (n, num_state, num_node) x (n, num_node, h*w) --> (n, num_state, h*w)
        x_state_reshaped = torch.matmul(x_n_rel, x_rproj_reshaped)

        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

        # (n, num_state, h*w) --> (n, num_state, h, w)
        x_state = x_state_reshaped.view(n, self.num_s, *x.size()[2:])

        # -----------------
        # (n, num_state, h, w) -> (n, num_in, h, w)
        out = x + self.blocker(self.conv_extend(x_state))

        return out


class GloRe_Unit_1D(GloRe_Unit):
    def __init__(self, num_in, num_mid, normalize=False):
        """
        Set 'normalize = True' if the input size is not fixed
        """
        super(GloRe_Unit_1D, self).__init__(num_in, num_mid,
                                            ConvNd=nn.Conv1d,
                                            BatchNormNd=nn.BatchNorm1d,
                                            normalize=normalize)

class GloRe_Unit_2D(GloRe_Unit):
    def __init__(self, num_in, num_mid, normalize=False):
        """
        Set 'normalize = True' if the input size is not fixed
        """
        super(GloRe_Unit_2D, self).__init__(num_in, num_mid,
                                            ConvNd=nn.Conv2d,
                                            BatchNormNd=nn.BatchNorm2d,
                                            normalize=normalize)

class GloRe_Unit_3D(GloRe_Unit):
    def __init__(self, num_in, num_mid, normalize=False):
        """
        Set 'normalize = True' if the input size is not fixed
        """
        super(GloRe_Unit_3D, self).__init__(num_in, num_mid,
                                            ConvNd=nn.Conv3d,
                                            BatchNormNd=nn.BatchNorm3d,
                                            normalize=normalize)


if __name__ == '__main__':

    for normalize in [True, False]:
        data = torch.autograd.Variable(torch.randn(2, 32, 1))
        net = GloRe_Unit_1D(32, 16, normalize)
        print(net(data).size())

        data = torch.autograd.Variable(torch.randn(2, 32, 14, 14))
        net = GloRe_Unit_2D(32, 16, normalize)
        print(net(data).size())

        data = torch.autograd.Variable(torch.randn(2, 32, 8, 14, 14))
        net = GloRe_Unit_3D(32, 16, normalize)
        print(net(data).size())


# -*- coding: utf-8 -*-
"""Functions for controlling the family-wise error rate (FWER).
This program code is part of the MultiPy (Multiple Hypothesis Testing in
Python) package.
Author: Tuomas Puoliväli (tuomas.puolivali@helsinki.fi)
Last modified: 27th December 2017.
License: Revised 3-clause BSD
Source: https://github.com/puolival/multipy/blob/master/fwer.py
References:
[1] Hochberg Y (1988): A sharper Bonferroni procedure for multiple tests of
    significance. Biometrika 75(4):800-802.
[2] Holm S (1979): A simple sequentially rejective multiple test procedure.
    Scandinavian Journal of Statistics 6(2):65-70.
[3] Sidak Z (1967): Confidence regions for the means of multivariate normal
    distributions. Journal of the American Statistical Association 62(318):
    626-633.
WARNING: These functions have not been entirely validated yet.
"""

import numpy as np

def bonferroni(pvals, alpha=0.05):
    """A function for controlling the FWER at some level alpha using the
    classical Bonferroni procedure.
    Input arguments:
    pvals       - P-values corresponding to a family of hypotheses.
    alpha       - The desired family-wise error rate.
    Output arguments:
    significant - An array of flags indicating which p-values are significant
                  after correcting for multiple comparisons.
    """
    m, pvals = len(pvals), np.asarray(pvals)
    return pvals < alpha/float(m)

def hochberg(pvals, alpha=0.05):
    """A function for controlling the FWER using Hochberg's procedure [1].
    Input arguments:
    pvals       - P-values corresponding to a family of hypotheses.
    alpha       - The desired family-wise error rate.
    Output arguments:
    significant - An array of flags indicating which p-values are significant
                  after correcting for multiple comparisons.
    """
    m, pvals = len(pvals), np.asarray(pvals)
    # Sort the p-values into ascending order
    ind = np.argsort(pvals)

    """Here we have k+1 (and not just k) since Python uses zero-based
    indexing."""
    test = [p <= alpha/(m+1-(k+1)) for k, p in enumerate(pvals[ind])]
    significant = np.zeros(np.shape(pvals), dtype='bool')
    significant[ind[0:np.sum(test)]] = True
    return significant

def holm_bonferroni(pvals, alpha=0.05):
    """A function for controlling the FWER using the Holm-Bonferroni
    procedure [2].
    Input arguments:
    pvals       - P-values corresponding to a family of hypotheses.
    alpha       - The desired family-wise error rate.
    Output arguments:
    significant - An array of flags indicating which p-values are significant
                  after correcting for multiple comparisons.
    """
    m, pvals = len(pvals), np.asarray(pvals)
    ind = np.argsort(pvals)
    test = [p > alpha/(m+1-k) for k, p in enumerate(pvals[ind])]

    """The minimal index k is m-np.sum(test)+1 and the hypotheses 1, ..., k-1
    are rejected. Hence m-np.sum(test) gives the correct number."""
    significant = np.zeros(np.shape(pvals), dtype='bool')
    significant[ind[0:m-np.sum(test)]] = True
    return significant

def sidak(pvals, alpha=0.05):
    """A function for controlling the FWER at some level alpha using the
    procedure by Sidak [3].
    Input arguments:
    pvals       - P-values corresponding to a family of hypotheses.
    alpha       - The desired family-wise error rate.
    Output arguments:
    significant - An array of flags indicating which p-values are significant
                  after correcting for multiple comparisons.
    """
    n, pvals = len(pvals), np.asarray(pvals)
    return pvals < 1. - (1.-alpha) ** (1./n)


from scipy.interpolate import UnivariateSpline

def _ecdf(x):
    '''no frills empirical cdf used in fdrcorrection
    '''
    nobs = len(x)
    return np.arange(1,nobs+1)/float(nobs)

multitest_methods_names = {'b': 'Bonferroni',
                           's': 'Sidak',
                           'h': 'Holm',
                           'hs': 'Holm-Sidak',
                           'sh': 'Simes-Hochberg',
                           'ho': 'Hommel',
                           'fdr_bh': 'FDR Benjamini-Hochberg',
                           'fdr_by': 'FDR Benjamini-Yekutieli',
                           'fdr_tsbh': 'FDR 2-stage Benjamini-Hochberg',
                           'fdr_tsbky': 'FDR 2-stage Benjamini-Krieger-Yekutieli',
                           'fdr_gbs': 'FDR adaptive Gavrilov-Benjamini-Sarkar'
                           }

_alias_list = [['b', 'bonf', 'bonferroni'],
               ['s', 'sidak'],
               ['h', 'holm'],
               ['hs', 'holm-sidak'],
               ['sh', 'simes-hochberg'],
               ['ho', 'hommel'],
               ['fdr_bh', 'fdr_i', 'fdr_p', 'fdri', 'fdrp'],
               ['fdr_by', 'fdr_n', 'fdr_c', 'fdrn', 'fdrcorr'],
               ['fdr_tsbh', 'fdr_2sbh'],
               ['fdr_tsbky', 'fdr_2sbky', 'fdr_twostage'],
               ['fdr_gbs']
               ]


multitest_alias = {}
for m in _alias_list:
    multitest_alias[m[0]] = m[0]
    for a in m[1:]:
        multitest_alias[a] = m[0]


def lsu(pvals, alpha=0.05,method='indep'):
    '''pvalue correction for false discovery rate

    This covers Benjamini/Hochberg for independent or positively correlated and
    Benjamini/Yekutieli for general or negatively correlated tests. Both are
    available in the function multipletests, as method=`fdr_bh`, resp. `fdr_by`.

    Parameters
    ----------
    pvals : array_like
        set of p-values of the individual tests.
    alpha : float
        error rate
    method : {'indep', 'negcorr'}
    is_sorted : bool
        If False (default), the p_values will be sorted, but the corrected
        pvalues are in the original order. If True, then it assumed that the
        pvalues are already sorted in ascending order.

    Returns
    -------
    rejected : ndarray, bool
        True if a hypothesis is rejected, False if not
    pvalue-corrected : ndarray
        pvalues adjusted for multiple hypothesis testing to limit FDR

    Notes
    -----

    If there is prior information on the fraction of true hypothesis, then alpha
    should be set to alpha * m/m_0 where m is the number of tests,
    given by the p-values, and m_0 is an estimate of the true hypothesis.
    (see Benjamini, Krieger and Yekuteli)

    The two-step method of Benjamini, Krieger and Yekutiel that estimates the number
    of false hypotheses will be available (soon).

    Method names can be abbreviated to first letter, 'i' or 'p' for fdr_bh and 'n' for
    fdr_by.
    '''
    pvals = np.asarray(pvals)
    is_sorted=False
    if not is_sorted:
        pvals_sortind = np.argsort(pvals)
        pvals_sorted = np.take(pvals, pvals_sortind)
    else:
        pvals_sorted = pvals  # alias

    if method in ['i', 'indep', 'p', 'poscorr']:
        ecdffactor = _ecdf(pvals_sorted)
    elif method in ['n', 'negcorr']:
        cm = np.sum(1./np.arange(1, len(pvals_sorted)+1))   #corrected this
        ecdffactor = _ecdf(pvals_sorted) / cm
##    elif method in ['n', 'negcorr']:
##        cm = np.sum(np.arange(len(pvals)))
##        ecdffactor = ecdf(pvals_sorted)/cm
    else:
        raise ValueError('only indep and negcorr implemented')
    reject = pvals_sorted <= ecdffactor*alpha
    if reject.any():
        rejectmax = max(np.nonzero(reject)[0])
        reject[:rejectmax] = True

    pvals_corrected_raw = pvals_sorted / ecdffactor
    pvals_corrected = np.minimum.accumulate(pvals_corrected_raw[::-1])[::-1]
    del pvals_corrected_raw
    pvals_corrected[pvals_corrected>1] = 1
    if not is_sorted:
        pvals_corrected_ = np.empty_like(pvals_corrected)
        pvals_corrected_[pvals_sortind] = pvals_corrected
        del pvals_corrected
        reject_ = np.empty_like(reject)
        reject_[pvals_sortind] = reject
        return reject_, pvals_corrected_
    else:
        return reject, pvals_corrected



from math import sqrt
from NaroNet.utils.utilz import auc_roc_curve
from NaroNet.utils.utilz import confusion_matrix

def confidence_interval_accuracy(accuracy,n_samples,Experiment_Name):
    z=1.96 # in 95% confidence interval
    interval = z * sqrt( (accuracy * (1 - accuracy)) / n_samples)
    print(Experiment_Name+', Accuracy is: {}%+-{}'.format(accuracy,interval))

# seed as early as possible

import numpy as np
np.random.seed(1234)
rng=np.random.RandomState(1234)



# Calculate confidence_interval_accuracy
confidence_interval_accuracy(accuracy=0.9167,n_samples=12,Experiment_Name='POLE subject-wise')
confidence_interval_accuracy(accuracy=0.9673,n_samples=336,Experiment_Name='POLE image-wise')

# Calculate AUC with confidence interval
# confidence_interval_AUC(y_true, y_pred)


# Recalculate roc curve for pairs of labels.
import pandas as pd
import numpy as np
process_dir = '/gpu-data/djsanchez/Images-ZuriBaselRisk_v4/NaroNet/Risk_3_classes_OSmonth/Cross_validation_results/'
Image_preds = pd.read_excel(process_dir+'Prediction_values_Risk_3_classes_Fold10.xlsx')
subject_preds = []
subject_labels = []
for S_N in np.unique(Image_preds['Subject_name']):
    subject_labels.append(list(Image_preds['Label'][np.array(Image_preds['Subject_name']==S_N)])[0])
    Image_predictions_str = list(Image_preds['Prediction'][np.array(Image_preds['Subject_name']==S_N)])
    Image_predictions=[]
    for i_p_s in Image_predictions_str:
        Image_predictions.append([float(i) for i in i_p_s.split('[')[1].split(']')[0].split(' ') if not ''==i ])
    subject_preds.append(np.stack(Image_predictions).mean(0))

thresholds,fpr,tpr = auc_roc_curve(['R2','R1','R3'],np.stack(subject_labels),
                   np.stack(subject_preds),process_dir,'POLEE','subject_wise')                              
confusion_matrix(['R2','R1','R3'],np.stack(subject_labels),np.stack(subject_preds),
                process_dir,thresholds,fpr,tpr,'POLEE','subject_wise')




import torch


def euclidean_norm_squared(vec_list):
    return sum(torch.sum(v ** 2).item() for v in vec_list)


def modify_lr(optimizer, mul):
    for param_group in optimizer.param_groups:
        param_group['lr'] *= mul


def get_lr(optimizer):
    lrs = [param_group['lr'] for param_group in optimizer.param_groups]
    for lr in lrs:
        if lr != lrs[0]:
            return lrs
    return lrs[0]


DENOMINATOR_EPS = 1e-25


class LossgradOptimizer:

    def __init__(self, optimizer, net, criterion, c=1.05):
        self.optimizer = optimizer
        self.c = c
        self.net = net
        self.criterion = criterion

    def load_state_dict(self, state_dict):
        if 'c' in state_dict:
            self.c = state_dict['c']
            del state_dict['c']
        self.optimizer.load_state_dict(state_dict)

    def state_dict(self):
        state_dict = self.optimizer.state_dict()
        state_dict['c'] = self.c
        return state_dict

    def step(self, X, y, loss):
        with torch.no_grad():
            grad_norm_squared = euclidean_norm_squared((-p.grad for p
                                                        in self.net.parameters()))
            lred = grad_norm_squared * get_lr(self.optimizer)
            approx = loss.item() - lred
            self.optimizer.step()
            actual = self.criterion(self.net(X), y).item()
            rel_err = (actual - approx) / (lred + DENOMINATOR_EPS)
        if rel_err > 0.5:
            h_mul = 1 / self.c
        else:
            h_mul = self.c
        modify_lr(self.optimizer, h_mul)
        return rel_err, grad_norm_squared

    def get_lr(self):
        return get_lr(self.optimizer)

class LossgradOptimizer:

    def __init__(self, optimizer, net, criterion, c=1.05):
        self.optimizer = optimizer
        self.c = c
        self.net = net
        self.criterion = criterion

    def load_state_dict(self, state_dict):
        if 'c' in state_dict:
            self.c = state_dict['c']
            del state_dict['c']
        self.optimizer.load_state_dict(state_dict)

    def state_dict(self):
        state_dict = self.optimizer.state_dict()
        state_dict['c'] = self.c
        return state_dict

    def step(self, X, y, loss):
        with torch.no_grad():
            grad_norm_squared = euclidean_norm_squared((-p.grad for p
                                                        in self.net.parameters()))
            lred = grad_norm_squared * get_lr(self.optimizer)
            approx = loss.item() - lred
            self.optimizer.step()
            actual = self.criterion(self.net(X), y).item()
            rel_err = (actual - approx) / (lred + DENOMINATOR_EPS)
        if rel_err > 0.5:
            h_mul = 1 / self.c
        else:
            h_mul = self.c
        modify_lr(self.optimizer, h_mul)
        return rel_err, grad_norm_squared

    def get_lr(self):
        return get_lr(self.optimizer)


import os 
import pandas as pd
import json
import numpy as np
import copy
import seaborn as sns
from NaroNet.BioInsights.add_annotation_stat import add_stat_annotation
import matplotlib.pyplot as plt
import os 
import pandas as pd
import json
import numpy as np
import copy
import seaborn as snss
import matplotlib.pyplot as plt
from itertools import combinations

def perf_by_pairs(save_dir, topk, file_name, frame_info):
    """
    Heatmap showing the performance by pairs.
    """
    SelectedCombinations = ['GLORE=True','ClusteringOrAttention=True','1cell1cluster=True','ortho=True','min_Cell_entropy=True','MinCut=True','Max_Pat_Entropy=True','Lasso_Feat_Selection=True']
    frame_Interp_Cop = []
    for fpa_1st in frame_info:
        for fpa_2nd in frame_info:  
            # In case those two consist of a combination of interesting modules.  
            if (fpa_1st in SelectedCombinations) and (fpa_2nd in SelectedCombinations):
                PosPairs = [frame_info[fpa_1st][idx] for idx in range(len(frame_info[fpa_1st])) if (frame_info[fpa_1st][idx]!=-1000 and frame_info[fpa_2nd][idx]!=-1000)]
                PosPairs.sort()
                if 'Interp' in file_name:
                    frame_Interp_Cop.append(PosPairs[-topk:])
                else:
                    frame_Interp_Cop.append(PosPairs[:topk])
    # Generate Heatmap
    heatmap = np.zeros((len(SelectedCombinations),len(SelectedCombinations)))
    n_iter = 0
    for n_1st, comb_1st in enumerate(SelectedCombinations):
        for n_2nd, comb_2nd in enumerate(SelectedCombinations):
            heatmap[n_1st,n_2nd] = sum(frame_Interp_Cop[n_iter])/topk
            n_iter+=1
    plt.close()
    ax = sns.heatmap(heatmap, annot=True)
    ax.set_xticklabels(['Global reasoning unit','Softmax activation function','Max-pooling','Orthogonal Loss','Patch entropy loss', 'Mincut loss','Patient entropy loss','Lasso loss (last layers)'],rotation=90)
    ax.set_yticklabels(['Global reasoning unit','Softmax activation function','Max-pooling','Orthogonal Loss','Patch entropy loss', 'Mincut loss','Patient entropy loss','Lasso loss (last layers)'],rotation=0)
    plt.tight_layout()
    plt.savefig(save_dir+file_name+'_heatmapPairs.png',dpi=600)

def showBoxplot(save_dir,pandasDF,parameters,ylabel,file_name):
    '''
        Show boxplot
    '''
    parameters = sorted(parameters)
    plt.close()
    ax = sns.boxplot(data=pandasDF, order=parameters)
    sns.swarmplot(data=pandasDF, order=parameters, color='0.25')
    ax.tick_params(labelsize=10)
    ax.set_xticklabels([p.split('=')[1]for p in parameters], fontsize=18)
    ax.set_ylabel(ylabel, fontsize=18)
    ax.set_title(parameters[0].split('=')[0], fontsize=18)
    # test_results = add_stat_annotation(ax, data=pandasDF, x=None, y=None, order=parameters,
    #                                 box_pairs=list(combinations(parameters,2)),
    #                                 test='Mann-Whitney', text_format='star',
    #                                 loc='inside', verbose=2)
    plt.savefig(save_dir+file_name+'_'+parameters[0].split('=')[0]+'_stats.png',dpi=600)

def obtain_configuration_names(load_dir):
    '''
        Obtain configuration and save it in frame_PredAcc
    '''
    
    # Initialize configruration 
    frame_PredAcc = {}
    init_list = list(-1000*np.ones(len(os.listdir(load_dir))))

    # Traverse load_dir    
    for n_run, dname in enumerate(os.listdir(load_dir)):
        
        # Load only json result files.
        if ('.json' in dname) or ('.xlsx' in dname) or ('.png' in dname) or ('.tmp' in dname) or ('.pkl' in dname):
            continue

        # Open Json Results file    
        with open(load_dir+dname+'/result.json', 'r') as json_file:        
            
            # Read each line from the results file.
            lines = json_file.readlines()
            if len(lines)>0: # Check if runs were finished
                data = json.loads(lines[-1])                     
                
                # Obtain names of the configuration (e.g. "1cell1cluster=True")
                for d in data['config']:
                    frame_PredAcc[d+'='+str(data['config'][d])] = copy.deepcopy(init_list)
    frame_Interp = copy.deepcopy(frame_PredAcc)

    return frame_PredAcc, frame_Interp

def obtain_configuration_performance(load_dir,frame_PredAcc,frame_Interp,metric):
    '''
        Assign performance to configurations
    '''

    # Traverse load_dir    
    for n_run, dname in enumerate(os.listdir(load_dir)):
        
        # Load only json result files.
        if ('.json' in dname) or ('.xlsx' in dname) or ('.png' in dname) or ('.tmp' in dname) or ('.pkl' in dname):
            continue
        
        # Open Json Results file  
        with open(load_dir+dname+'/result.json', 'r') as json_file:  
            lines = json_file.readlines()  
            # Check if runs were finished
            if len(lines)>0:  
                data = json.loads(lines[-1])  
                # For each possible configuration check if this run has been chosen  
                for fr in frame_PredAcc:  
                    # has this configuration been chosen by the run?  
                    if fr.split('=')[1] == str(data['config'][fr.split('=')[0]]):  
                        
                        try:# Assign the val loss  
                            frame_PredAcc[fr][n_run]=data[metric]  
                            if 'Synthetic' in load_dir:  
                                frame_Interp[fr][n_run]=data['interpretability']*100  
                        except:
                            continue
    return frame_PredAcc, frame_Interp  

def save_architecture_search_stats(save_dir,load_dir,topk):
    '''
        Calculate statistics showing which NaroNet architectures perform best
    '''
    
    # Load configurations
    frame_PredAcc, frame_Interp = obtain_configuration_names(load_dir)
    frame_PredAcc, frame_Interp = obtain_configuration_performance(load_dir,frame_PredAcc,frame_Interp,'cross_ent_test')

    # Save run configurations to excel (just in case)
    pd.DataFrame.from_dict(frame_PredAcc).to_excel(load_dir+'/predAcc.xlsx')

    # Obtain performance by pairs of parameters.
    # perf_by_pairs(save_dir, topk, 'Accuracy', frame_PredAcc)
    # if 'Synthetic' in load_dir:
    #     perf_by_pairs(save_dir,topk, 'Interp', frame_Interp)

    # Take best performing architectures per parameters.
    conf_param_acc, topk_acc = topk_performing_architectures(frame_PredAcc, mode='min',topk=topk)    
    if 'Synthetic' in load_dir:
        conf_param_int, topk_int = topk_performing_architectures(frame_Interp, mode='min',topk=topk)
            
    # Display boxplots showing the performance 
    for p in conf_param_acc:
        showBoxplot(save_dir,pd.DataFrame.from_dict(topk_acc),p,ylabel='Cross-validation test loss',file_name='Acc')
    if 'Synthetic' in load_dir:
        for p in conf_param_int:
            showBoxplot(save_dir,pd.DataFrame.from_dict(topk_int),p,ylabel='Interpretability Accuracy (%)',file_name='Interp')

def topk_performing_architectures(frame, mode,topk):
    '''
        Obtain Best performing values for each test type with respect the interpretability
    '''
    
    # Initialize
    frame_topk = {}

    # Take the topk performing values.
    for fpa in frame:    
        listt = [i for i in frame[fpa] if i!=-1000]
        listt.sort()
        if len(listt)>=topk:        
            if mode=='min':
                frame_topk[fpa] = [i if i!=-1 else 0 for i in listt[:topk]]
            elif mode=='max':
                frame_topk[fpa] = [i if i!=-1 else 0 for i in listt[-topk:]]

    # List configurations that were chosen.
    parameters_all = []
    for fpa_init in frame_topk:
        fpa_init_type = fpa_init.split('=')[0]
        parameters = []
        for fpa in frame_topk:
            if fpa.split('=')[0]==fpa_init_type:
                parameters.append(fpa)
        if len(parameters)>1:
            parameters_all.append(parameters)
    
    return parameters_all, frame_topk


def extract_best_result(load_dir,metric,best_params):
    '''
        Method  that extracts the best result out of all the executions 
    '''
    n_runs = 0
    saved_metric = 0 if metric == 'acc_test' else 10

    # Assign runs performance to each parameter.
    for n_run, dname in enumerate(os.listdir(load_dir)):

        if ('.json' in dname) or ('.xlsx' in dname) or ('.png' in dname) or ('.pkl' in dname) or ('.tmp' in dname):
            continue

        # Open Json Results file    
        with open(load_dir+'/'+dname+'/result.json', 'r') as json_file:            
            
            lines = json_file.readlines()
            if len(lines)>0: # Check if runs were finished
                n_runs +=1
                data = json.loads(lines[-1])         
                # print(str(data[metric]))
                try:
                    if saved_metric>data[metric]:                
                        # print(metric+": "+str(data[metric])+'\ntrain_Cross_entropy:'+str(data['train_Cross_entropy'])+'\nacc_test:'+str(data['acc_test'])+'\ntrain_acc:'+str(data['train_acc']))
                        saved_metric = data[metric]
                        config = data['config']
                except:
                    continue
    
    if 'config' in locals():
        return config, n_runs
    else:
        return best_params, n_runs


if __name__ == "__main__":
    load_dir = '/home/djsanchez/ray_results/ExplainableML/'
    save_architecture_search_stats(load_dir,load_dir,10)


import os
import numpy as np
import pandas as pd

BaselData = pd.read_excel('/gpu-data/djsanchez/Images-ZuriBaselRisk_1Im_Cell_v4/Raw_Data/Experiment_Information/Patient_to_Image.xlsx')
output_folder = '/gpu-data/djsanchez/Images-ZuriBaselRisk_1Im_Cell_v4/Patch_Contrastive_Learning/Image_Patch_Representation/'
input_folder = '/gpu-data/djsanchez/Images-ZuriBaselRisk_1Im_Cell_v4/Patch_Contrastive_Learning/Image_Patch_Representation(1Im)/'
images = os.listdir('/gpu-data/djsanchez/Images-ZuriBaselRisk_1Im_Cell_v4/Patch_Contrastive_Learning/Image_Patch_Representation(1Im)/')

# Iterate over images
for n_file in range(len(images)): 
    index = list(BaselData['Image_Name']).index(images[n_file][:-3]+'tiff')
    subject_name = BaselData['Subject_Name'][index]
    if str(subject_name)+'.npy' in os.listdir(output_folder):
        previous_patient = np.load(output_folder+str(subject_name)+'.npy')
        image = np.load(input_folder+images[n_file]) + previous_patient[:,[0,1]].max() + 100
        updated_patient = np.concatenate((previous_patient,image),axis=0)
        np.save(output_folder+str(subject_name)+".npy",updated_patient)
    else:
        np.save(output_folder+str(subject_name)+".npy",np.load(input_folder+images[n_file]))  



import pandas as pd
import os
import numpy as np

def load_table(path):
    return pd.read_excel(path)

def meanerror_predictions_and_labels(image_names,predictions,labels,risk_group):
    meanerror = []
    for n_image, image_label in enumerate(labels):
        if image_label>risk_group[0] and image_label<risk_group[1]:
            meanerror.append(abs(image_label-max(float(predictions[n_image].split('[')[1].split(']')[0]),0)))
    return np.array(meanerror).mean(), np.array(meanerror).std()



table = load_table('/gpu-data/djsanchez/Images-ZuriBaselRisk_1Im_v4/NaroNet/OSmonth_Risk_3_classes/Cross_validation_results/Prediction_values_OSmonth_Fold10.xlsx')

image_names = table['Subject_name']
predictions = table['Prediction']
labels = table['Label']

intervals = [(0,53),(54,119),(120,210)]

for risk_group in intervals:
    meanerror, stderror = meanerror_predictions_and_labels(image_names,predictions,labels,risk_group) 
    print('Risk Group: '+str(risk_group)+'_Mean errror: '+str(meanerror)+'+-'+str(stderror))






import argparse
import numpy as np
from hyperopt import hp
import torch

def parameters(path, debug):

    # args=DefaultParameters(path)
    args={}
    args['path'] = path

    # Use optimized parameters depending on the experiment.
    if 'STK11_WholeSlide' in path:            
        # Patch contrastive learning parameters
        args['PCL_Out_Dimensions'] = 256
        args['PCL_batch_size']=12
        args['PCL_epochs']=400
        args['PCL_patch_size']=60
        args['PCL_alpha_L']=1.1#1.07 # The value of alpha_L in the manuscript
        args['PCL_ZscoreNormalization']=True        
        args['PCL_width_CNN']=1 # [1, 2, 4]           
        args['PCL_depth_CNN']=18#4 # [1, 2, 4] 
        args['PCL_pheno_neigh']=False # True trains the model for neighborhoods and phenotypes

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['Mutation'] 

        # Optimization Parameters
        args['verbose'] = 1 # IF zero nothing, if 1 show information on console
        args['num_samples_architecture_search'] = 40
        args['epochs'] = 15# if debug else hp.quniform('epochs', 5, 25, 1)
        args['epoch'] = 0
        args['lr_decay_factor'] = 0.5# if debug else hp.uniform('lr_decay_factor', 0, 0.75)
        args['lr_decay_step_size'] = max(int(args['epochs']/1),1)# if debug else hp.quniform('lr_decay_step_size', 2, 20, 1)        
        args['weight_decay'] = 2 if debug=='Index' else hp.choice('weight_decay',[0.1,0.01,0.001,0.0001]) if debug=='Object' else 0.01
        args['batch_size'] = 2 if debug=='Index' else hp.choice('batch_size', [8,16,32,64,128,256]) if debug=='Object' else 32
        args['lr'] = 1 if debug=='Index' else hp.choice('lr', [0.1,0.01,0.001]) if debug=='Object' else 0.01
        args['useOptimizer'] = 'ADAM' #0 if debug else hp.choice("useOptimizer", ['ADAM', 'ADAMW', 'ADABound']) # 0:ADAM, 1:ADAMW, 2:ADABound

        # General
        args['context_size'] = 15#0 if debug else hp.choice("context_size", [15])
        args['num_classes'] = 3
        args['MultiClass_Classification']=1
        args['showHowNetworkIsTraining'] = False # Creates a GIF of the learning clusters!
        args['visualizeClusters'] = True
        args['learnSupvsdClust'] = True
        args['recalculate'] = False
        args['is_nested_cross_validation'] = False # Whether to use nested cross validation or cross validation.
        args['folds'] = 5
        args['device'] = 'cuda:0'
        args['normalizeFeats'] = 0 if debug=='Index' else hp.choice("normalizeFeats", [True,False]) if debug=='Object' else True
        args['normalizeCells'] = 0 if debug=='Index' else hp.choice("normalizeCells", [True,False]) if debug=='Object' else True 
        args['Batch_Normalization'] = 1 if debug=='Index' else hp.choice("Batch_Normalization", [True,False]) if debug=='Object' else False
        args['normalizePercentile'] = False#1 if debug else hp.choice("normalizePercentile", [True,False])
        args['dataAugmentationPerc'] = 0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['hiddens'] = 1 if debug=='Index' else hp.choice('hiddens', [32,44,64,86,128]) if debug=='Object' else 44                
        args['clusters1'] = 2 if debug=='Index' else hp.choice('clusters1',[4,7,10]) if debug=='Object' else 10        
        args['clusters2'] = 2 if debug=='Index' else hp.choice('clusters2',[3,6,9]) if debug=='Object' else 9 
        args['clusters3'] = 1 if debug=='Index' else hp.choice('clusters3',[2,7,8]) if debug=='Object' else 7          
        args['LSTM'] = False#0 if debug else hp.choice("LSTM", [True,False])
        args['GLORE'] = False#1 if debug=='Index' else hp.choice('GLORE',[True,False]) if debug=='Object' else False 
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = True  
        args['Area_Learning'] = True
        args['DeepSimple'] = True
        args['isAttentionLayer'] = False#1 if debug else hp.choice("isAttentionLayer", [True,False])
        args['ClusteringOrAttention'] = True#0 if debug=='Index' else hp.choice("ClusteringOrAttention", [True,False]) if debug=='Object' else True        
        args['1cell1cluster'] = False#1 if debug=='Index' else hp.choice("1cell1cluster", [True,False]) if debug=='Object' else False                
        args['dropoutRate'] = 1 if debug=='Index' else hp.choice('dropoutRate', [0.1, 0.2, 0.3,0.4]) if debug=='Object' else 0.2        
        args['AttntnSparsenss'] = False#1 if debug else hp.choice("AttntnSparsenss", [True,False])
        args['attntnThreshold'] = 0# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
        args['GraphConvolution'] = 'ResNet'#0 if debug=='Index' else hp.choice('GraphConvolution', ['ResNet', 'IncepNet', 'JKNet']) if debug=='Object' else 'ResNet'                
        args['n-hops'] = 1 if debug=='Index' else hp.choice('n-hops', [2, 3]) if debug=='Object' else 3                        
        args['modeltype'] = 'SAGE'#0 if debug else hp.choice('modeltype', ['SAGE', 'SGC']) # 0:SAGE, 1:SGC
        args['ObjectiveCluster'] = True#1 if debug else hp.choice('ObjectiveCluster', [True, False]) # Whether to learn a X and S embedding or just the clustering
        args['ReadoutFunction'] = False#0 if debug else hp.choice('ReadoutFunction', ['MAX', 'SUM', 'DeepSets']) # Choose the readout function    
        args['NearestNeighborClassification'] = False#1 if debug else hp.choice('NearestNeighborClassification', [True, False]) # Use the nearest Neighbor strategy
        args['NearestNeighborClassification_Lambda0'] = 1#0 if debug else hp.choice("NearestNeighborClassification_Lambda0", [0.1, 0.01 ,0.001, 0.0001])
        args['NearestNeighborClassification_Lambda1'] = 1#0 if debug else hp.choice("NearestNeighborClassification_Lambda1", [0.1, 0.01, 0.001, 0.0001])
        args['NearestNeighborClassification_Lambda2'] = 1#0 if debug else hp.choice("NearestNeighborClassification_Lambda2", [0.1, 0.01, 0.001, 0.0001])
        args['KinNearestNeighbors'] = 5#0 if debug else hp.choice('KinNearestNeighbors', [5, 10]) # Choose number of K in nearest neighbor strategy
        # Losses
        args['pearsonCoeffSUP'] = False#1 if debug else hp.choice("pearsonCoeffSUP", [True,False])
        args['pearsonCoeffUNSUP'] = False#1 if debug else hp.choice("pearsonCoeffUNSUP", [True,False])
        args['orthoColor'] = 0 if debug=='Index' else hp.choice("orthoColor", [True,False]) if debug=='Object' else True
        args['orthoColor_Lambda0'] = 0 if debug=='Index' else hp.choice("orthoColor_Lambda0", [0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.1                                
        args['orthoColor_Lambda1'] = 3 if debug=='Index' else hp.choice("orthoColor_Lambda1", [0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                              
        args['ortho'] = False#1 if debug=='Index' else hp.choice("ortho", [True,False]) if debug=='Object' else False                        
        args['ortho_Lambda0'] = 0# if debug=='Index' else hp.choice("ortho_Lambda0", [0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0.1                                
        args['ortho_Lambda1'] = 0#4 if debug=='Index' else hp.choice("ortho_Lambda1", [0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0                                
        args['ortho_Lambda2'] = 0#4 if debug=='Index' else hp.choice("ortho_Lambda2", [0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0                                
        args['min_Cell_entropy'] = True#0 if debug=='Index' else hp.choice("min_Cell_entropy", [True,False]) if debug=='Object' else True                                        
        args['min_Cell_entropy_Lambda0'] = 0 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda0", [1,0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 1   
        args['min_Cell_entropy_Lambda1'] = 1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda1", [1,0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0.1
        args['min_Cell_entropy_Lambda2'] = 1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda2", [1,0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0.1
        args['MinCut'] = False#0 if debug=='Index' else hp.choice("MinCut", [True,False]) if debug=='Object' else True        
        args['MinCut_Lambda0'] = 0#5 if debug=='Index' else hp.choice("MinCut_Lambda0", [1,0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0        
        args['MinCut_Lambda1'] = 0#1 if debug=='Index' else hp.choice("MinCut_Lambda1", [1,0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0.1        
        args['MinCut_Lambda2'] = 0#1 if debug=='Index' else hp.choice("MinCut_Lambda2", [1,0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0.1        
        args['F-test'] = False#1 if debug else hp.choice("F-test", [True,False])
        args['Max_Pat_Entropy'] = False#1 if debug=='Index' else hp.choice('Max_Pat_Entropy', [True, False]) if debug=='Object' else False                
        args['Max_Pat_Entropy_Lambda0'] = 0#4 if debug=='Index' else hp.choice("Max_Pat_Entropy_Lambda0", [1,0.1,0.01,0.001,0.0001]) if debug=='Object' else 0.0001        
        args['Max_Pat_Entropy_Lambda1'] = 0#1 if debug=='Index' else hp.choice("Max_Pat_Entropy_Lambda1", [1,0.1,0.01,0.001,0.0001]) if debug=='Object' else 0.1        
        args['Max_Pat_Entropy_Lambda2'] = 0#1 if debug=='Index' else hp.choice("Max_Pat_Entropy_Lambda2", [1,0.1,0.01,0.001,0.0001]) if debug=='Object' else 0.1        
        args['UnsupContrast'] = False#1 if debug else hp.choice("UnsupContrast", [True,False])
        args['UnsupContrast_Lambda0'] =0# 1 if debug else hp.choice("UnsupContrast_Lambda0", [1,0.1,0.01,0.001,0.0001])
        args['UnsupContrast_Lambda1'] =0# 1 if debug else hp.choice("UnsupContrast_Lambda1", [1,0.1,0.01,0.001,0.0001])
        args['UnsupContrast_Lambda2'] =0# 2 if debug else hp.choice("UnsupContrast_Lambda2", [1,0.1,0.01,0.001,0.0001])        
        args['Lasso_Feat_Selection'] = False#1 if debug=='Index' else hp.choice("Lasso_Feat_Selection", [True,False]) if debug=='Object' else False         
        args['Lasso_Feat_Selection_Lambda0'] = False#1 if debug=='Index' else hp.choice("Lasso_Feat_Selection_Lambda0", [1,0.1, 0.01,0.001,0]) if debug=='Object' else 0.1         
        args['SupervisedLearning_Lambda0'] = 0 if debug=='Index' else hp.choice("SupervisedLearning_Lambda0", [1,0.1, 0.01,0.001,0.0001]) if debug=='Object' else 1         
        args['SupervisedLearning_Lambda1'] = 0 if debug=='Index' else hp.choice("SupervisedLearning_Lambda1", [1,0.1, 0.01,0.001,0.0001]) if debug=='Object' else 1         
        args['SupervisedLearning_Lambda2'] = 0 if debug=='Index' else hp.choice("SupervisedLearning_Lambda2", [1,0.1, 0.01,0.001,0.0001]) if debug=='Object' else 1         
        args['SupervisedLearning_Lambda3'] = 0 if debug=='Index' else hp.choice("SupervisedLearning_Lambda3", [1,0.1, 0.01,0.001,0.0001]) if debug=='Object' else 1         
        args['SupervisedLearning'] = True# if debug else hp.choice("SupervisedLearning", [True,False])   

    elif 'STK11' in path:            
        
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # or resnet18
        args['PCL_N_Workers']=2 # Number of data loading workers 
        args['PCL_Epochs']=80 # Number of total epoch to run
        args['PCL_N_Crops_per_Image']=110 # Number of crops chosen per image
        args['PCL_Patch_Size']=30 # Size of patches in pixels
        args['PCL_Stride']=0 # Size of stride between consecutive patches
        args['PCL_Alpha_L']=1.15 # size ratio between image crop and image patch
        args['PCL_Z_Score']= True # Whether to apply z score per channel/marker
        args['PCL_Batch_Size']= 4 # Number of images per iteration
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of patch
        args['PCL_Learning_Rate']= 0.003 # Learning Rate
        args['PCL_Temperature']= 0.07 # Softmax temperature                
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # GPU where the training is carried out 

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['Mutation'] 

        # Architecture Search parameters
        args['num_samples_architecture_search'] = 800  

        # Optimization Parameters
        args['epochs'] = 3 if debug=='Index' else hp.choice('epochs',[6,15,30,100]) if debug=='Object' else 100
        args['lr_decay_factor'] = 0.25# if debug else hp.uniform('lr_decay_factor', 0, 0.75)
        args['lr_decay_step_size'] = 11000#max(int(args['epochs']/2),1)# if debug else hp.quniform('lr_decay_step_size', 2, 20, 1)        
        args['weight_decay'] = 4 if debug=='Index' else hp.choice('weight_decay',[0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0
        args['batch_size'] = 2 if debug=='Index' else hp.choice('batch_size', [2, 4, 8, 14, 24]) if debug=='Object' else 8
        args['lr'] = 0 if debug=='Index' else hp.choice('lr', [1,0.1,0.01,0.001,0.0001]) if debug=='Object' else 1

        # General
        args['training_MODE'] = 'TrainALL' # TrainALL, NestedCrossValidation, CrossValidation # Whether to use nested cross validation or cross validation.        
        args['folds'] = 8
        args['device'] = 'cuda:0'
        args['Batch_Normalization'] = True
        args['dataAugmentationPerc'] = 0 #0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['N_Phenotypes'] = 0 if debug=='Index' else hp.choice('clusters1',[8,10,15,20]) if debug=='Object' else 8     
        args['N_Neighborhoods'] = 1 if debug=='Index' else hp.choice('clusters2',[6,11,16,21]) if debug=='Object' else 11 
        args['N_Areas'] = 1 if debug=='Index' else hp.choice('clusters3',[2,4,7]) if debug=='Object' else 4      
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = True
        args['Area_Learning'] = True
        args['dropoutRate'] = 0 if debug=='Index' else hp.choice('dropoutRate', [0, 0.25]) if debug=='Object' else 0
        args['attntnThreshold'] = 0 if debug=='Index' else hp.choice('attntnThreshold', [0,.1,.2]) if debug=='Object' else 0 
        
        # Losses
        args['ContrastACC_Pheno'] = 2 if debug=='Index' else hp.choice("ContrastACC_Pheno", [1,0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0.01
        args['ContrastACC_Neigh'] = 0 if debug=='Index' else hp.choice("ContrastACC_Neigh", [1, 0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 1
        args['ContrastACC_Area'] = 4 if debug=='Index' else hp.choice("ContrastACC_Area", [1, 0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0.0001
        args['PatchEntropy_Pheno'] = 1 if debug=='Index' else hp.choice("PatchEntropy_Pheno", [1,0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0.1   
        args['PatchEntropy_Neigh'] = 1 if debug=='Index' else hp.choice("PatchEntropy_Neigh", [1,0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0.1
        args['PatchEntropy_Area'] = 4 if debug=='Index' else hp.choice("PatchEntropy_Area", [1,0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0.0001                                             

        # Bioinsights parameters
        args['Bio_save_Orig_im'] = False

    elif 'LF_Recurrence' in path:            
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # or resnet18
        args['PCL_N_Workers']=0 # Number of data loading workers 
        args['PCL_Epochs']=200 # Number of total epoch to run
        args['PCL_N_Crops_per_Image']=70 # Number of crops chosen per image
        args['PCL_Patch_Size']=30 # Size of patches in pixels
        args['PCL_Stride']=0 # Size of stride between consecutive patches
        args['PCL_Alpha_L']=1.15 # size ratio between image crop and image patch
        args['PCL_Z_Score']= True # Whether to apply z score per channel/marker 
        args['PCL_Batch_Size']= 5 # Number of images per iteration
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of patch
        args['PCL_Learning_Rate']= 0.0003 # Learning Rate
        args['PCL_Temperature']= 0.07 # Softmax temperature                
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # GPU where the training is carried out 

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['Relapse'] 

        # Architecture Search parameters
        args['num_samples_architecture_search'] = 700  

        # Optimization Parameters
        args['epochs'] = 80# if debug else hp.quniform('epochs', 5, 25, 1)
        args['lr_decay_factor'] = 0.1# if debug else hp.uniform('lr_decay_factor', 0, 0.75)
        args['lr_decay_step_size'] = 10000#max(int(args['epochs']/3),1)# if debug else hp.quniform('lr_decay_step_size', 2, 20, 1)        
        args['weight_decay'] = 0#3 if debug=='Index' else hp.choice('weight_decay',[0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0.0001
        args['batch_size'] = 4 if debug=='Index' else hp.choice('batch_size', [2, 4, 8, 16, 28]) if debug=='Object' else 28
        args['lr'] = 1 if debug=='Index' else hp.choice('lr', [1,0.5,0.01,0.001,0.0001]) if debug=='Object' else 0.5

        # General        
        args['training_MODE'] = 'TrainALL' # TrainALL, NestedCrossValidation, CrossValidation # Whether to use nested cross validation or cross validation.            
        args['folds'] = 1
        args['device'] = 'cuda:0'
        args['Batch_Normalization'] = True
        args['dataAugmentationPerc'] = 0 #0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['N_Phenotypes'] = 1 if debug=='Index' else hp.choice('clusters1',[5,8,15,20]) if debug=='Object' else 8     
        args['N_Neighborhoods'] = 2 if debug=='Index' else hp.choice('clusters2',[6,11,16,21]) if debug=='Object' else 16 
        args['N_Areas'] = 1 if debug=='Index' else hp.choice('clusters3',[2,4,7]) if debug=='Object' else 4      
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = True  
        args['Area_Learning'] = True
        args['dropoutRate'] = 0# 1 if debug=='Index' else hp.choice('dropoutRate', [0.1, 0.2, 0.3,0.4]) if debug=='Object' else 0.2        
        args['attntnThreshold'] = 0# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
        args['1Patch1Cluster'] = True# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  

        # Losses 
        args['ContrastACC_Pheno'] = 0#.001#4 if debug=='Index' else hp.choice("orthoColor_Lambda0", [1,0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                
        args['ContrastACC_Neigh'] = 0#.001#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['ContrastACC_Area'] = 0#.0001#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['PatchEntropy_Pheno'] = 0#.01#3 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda0", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.001   
        args['PatchEntropy_Neigh'] = 0#.01#4 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda1", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0
        args['PatchEntropy_Area'] = 0#.0001#1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda2", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.1                                  

        # Bioinsights parameters
        args['Bio_save_Orig_im'] = False


    elif 'LF_19H' in path:            
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # or resnet18
        args['PCL_N_Workers']=0 # Number of data loading workers 
        args['PCL_Epochs']=400 # Number of total epoch to run
        args['PCL_N_Crops_per_Image']=200 # Number of crops chosen per image
        args['PCL_Patch_Size']=30 # Size of patches in pixels
        args['PCL_Stride']=0 # Size of stride between consecutive patches
        args['PCL_Alpha_L']=1.15 # size ratio between image crop and image patch
        args['PCL_Z_Score']= True # Whether to apply z score per channel/marker 
        args['PCL_Batch_Size']= 6 # Number of images per iteration
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of patch
        args['PCL_Learning_Rate']= 0.0003 # Learning Rate
        args['PCL_Temperature']= 0.07 # Softmax temperature         
        args['PCL_eliminate_Black_Background'] = True
        args['PCL_eliminate_White_Background'] = False       
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # GPU where the training is carried out 
        args['PCL_Color_perturbation_Augmentation'] = True # Color augmentation for H&E tissue staining.

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['Relapse'] 

        # Architecture Search parameters
        args['num_samples_architecture_search'] = 700  

        # Optimization Parameters
        args['epochs'] = 8# if debug else hp.quniform('epochs', 5, 25, 1)
        args['lr_decay_factor'] = 0.1# if debug else hp.uniform('lr_decay_factor', 0, 0.75)
        args['lr_decay_step_size'] = 10000#max(int(args['epochs']/3),1)# if debug else hp.quniform('lr_decay_step_size', 2, 20, 1)        
        args['weight_decay'] = 0#3 if debug=='Index' else hp.choice('weight_decay',[0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0.0001
        args['batch_size'] = 26# if debug=='Index' else hp.choice('batch_size', [2, 4, 8, 16, 22]) if debug=='Object' else 22
        args['lr'] = 0.1#0 if debug=='Index' else hp.choice('lr', [1,0.1,0.01]) if debug=='Object' else 1

        # General        
        args['training_MODE'] = 'TrainALL' # TrainALL, NestedCrossValidation, CrossValidation # Whether to use nested cross validation or cross validation.            
        args['folds'] = 1
        args['device'] = 'cuda:0'
        args['Batch_Normalization'] = True
        args['dataAugmentationPerc'] = 0 #0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['N_Phenotypes'] = 1 if debug=='Index' else hp.choice('clusters1',[5,8,15,20]) if debug=='Object' else 8     
        args['N_Neighborhoods'] = 2 if debug=='Index' else hp.choice('clusters2',[6,11,16,21]) if debug=='Object' else 16 
        args['N_Areas'] = 1 if debug=='Index' else hp.choice('clusters3',[2,4,7]) if debug=='Object' else 4      
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = True  
        args['Area_Learning'] = True
        args['dropoutRate'] = 0# 1 if debug=='Index' else hp.choice('dropoutRate', [0.1, 0.2, 0.3,0.4]) if debug=='Object' else 0.2        
        args['attntnThreshold'] = 0# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
        args['1Patch1Cluster'] = False# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  


        # Losses
        args['ContrastACC_Pheno'] = 1#.001#4 if debug=='Index' else hp.choice("orthoColor_Lambda0", [1,0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                
        args['ContrastACC_Neigh'] = 1#.001#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['ContrastACC_Area'] = 1#.0001#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['PatchEntropy_Pheno'] = 0#.01#3 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda0", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.001   
        args['PatchEntropy_Neigh'] = 0#.01#4 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda1", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0
        args['PatchEntropy_Area'] = 0#.0001#1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda2", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.1                                  

        # Bioinsights parameters
        args['Bio_save_Orig_im'] = False

    elif 'Urothelial_ClinBen' in path:            
        
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # or resnet18
        args['PCL_N_Workers']=8 # Number of data loading workers 
        args['PCL_Epochs']=80 # Number of total epoch to run
        args['PCL_N_Crops_per_Image']=40 # Number of crops chosen per image
        args['PCL_Patch_Size']=30 # Size of patches in pixels
        args['PCL_Stride']=0 # Size of stride between consecutive patches
        args['PCL_Alpha_L']=1.15 # size ratio between image crop and image patch
        args['PCL_Z_Score']= True # Whether to apply z score per channel/marker
        args['PCL_Batch_Size']= 8 # Number of images per iteration
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of patch
        args['PCL_Learning_Rate']= 0.003 # Learning Rate
        args['PCL_Temperature']= 0.07 # Softmax temperature    
        args['PCL_eliminate_Black_Background'] = False
        args['PCL_eliminate_White_Background'] = False            
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # GPU where the training is carried out 

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['Clinical_Benefit'] 

        # Architecture Search parameters
        args['num_samples_architecture_search'] = 1500  

        # Optimization Parameters
        args['epochs'] = 1000# if debug else hp.quniform('epochs', 5, 25, 1)
        args['lr_decay_factor'] = 0.1# if debug else hp.uniform('lr_decay_factor', 0, 0.75)
        args['lr_decay_step_size'] = 10000#max(int(args['epochs']/5),1)# if debug else hp.quniform('lr_decay_step_size', 2, 20, 1)        
        args['weight_decay'] = 4 if debug=='Index' else hp.choice('weight_decay',[0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0
        args['batch_size'] = 3 if debug=='Index' else hp.choice('batch_size', [2, 4, 8, 24]) if debug=='Object' else 24
        args['lr'] = 0 if debug=='Index' else hp.choice('lr', [1,0.1,0.01,0.001,0.0001]) if debug=='Object' else 1

        # General
        args['training_MODE'] = 'TrainALL' # TrainALL, NestedCrossValidation, CrossValidation # Whether to use nested cross validation or cross validation.        
        args['folds'] = 1
        args['device'] = 'cuda:0'
        args['Batch_Normalization'] = True
        args['dataAugmentationPerc'] = 0 #0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['N_Phenotypes'] = 8#2 if debug=='Index' else hp.choice('clusters1',[5,10,15,20]) if debug=='Object' else 15     
        args['N_Neighborhoods'] = 16#1 if debug=='Index' else hp.choice('clusters2',[6,11,16,21]) if debug=='Object' else 11 
        args['N_Areas'] = 4#0 if debug=='Index' else hp.choice('clusters3',[2,4,7]) if debug=='Object' else 2      
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = True
        args['Area_Learning'] = True
        args['dropoutRate'] = 0# 1 if debug=='Index' else hp.choice('dropoutRate', [0.1, 0.2, 0.3,0.4]) if debug=='Object' else 0.2        
        args['attntnThreshold'] = 0# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
                
        # Losses
        args['ContrastACC_Pheno'] = 0.01#4 if debug=='Index' else hp.choice("orthoColor_Lambda0", [1,0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                
        args['ContrastACC_Neigh'] = 0.01#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['ContrastACC_Area'] = 0.0001#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['PatchEntropy_Pheno'] = 0.01#3 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda0", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.001   
        args['PatchEntropy_Neigh'] = 0.01#4 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda1", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0
        args['PatchEntropy_Area'] = 0.0001#1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda2", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.1          
        
        # Bioinsights parameters
        args['Bio_save_Orig_im'] = True

    elif 'Tumor_subtype' in path:            
        
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # or resnet18
        args['PCL_N_Workers']=8 # Number of data loading workers 
        args['PCL_Epochs']=80 # Number of total epoch to run
        args['PCL_N_Crops_per_Image']=40 # Number of crops chosen per image
        args['PCL_Patch_Size']=30 # Size of patches in pixels
        args['PCL_Stride']=0 # Size of stride between consecutive patches
        args['PCL_Alpha_L']=1.15 # size ratio between image crop and image patch
        args['PCL_Z_Score']= True # Whether to apply z score per channel/marker
        args['PCL_Batch_Size']= 8 # Number of images per iteration
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of patch
        args['PCL_Learning_Rate']= 0.003 # Learning Rate
        args['PCL_Temperature']= 0.07 # Softmax temperature                
        args['PCL_eliminate_Black_Background'] = False
        args['PCL_eliminate_White_Background'] = False
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # GPU where the training is carried out 

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['Clinical_Benefit'] 

        # Architecture Search parameters
        args['num_samples_architecture_search'] = 1500  

        # Optimization Parameters
        args['epochs'] = 500
        args['lr_decay_factor'] = 0.1
        args['lr_decay_step_size'] = 10000
        args['weight_decay'] = 0
        args['batch_size'] = 24
        args['lr'] = 1 

        # General
        args['training_MODE'] = 'TrainALL' # TrainALL, NestedCrossValidation, CrossValidation # Whether to use nested cross validation or cross validation.        
        args['folds'] = 1
        args['device'] = 'cuda:0'
        args['Batch_Normalization'] = True
        args['dataAugmentationPerc'] = 0 #0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['N_Phenotypes'] = 8#2 if debug=='Index' else hp.choice('clusters1',[5,10,15,20]) if debug=='Object' else 15     
        args['N_Neighborhoods'] = 16#1 if debug=='Index' else hp.choice('clusters2',[6,11,16,21]) if debug=='Object' else 11 
        args['N_Areas'] = 4#0 if debug=='Index' else hp.choice('clusters3',[2,4,7]) if debug=='Object' else 2      
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = True
        args['Area_Learning'] = True
        args['dropoutRate'] = 0# 1 if debug=='Index' else hp.choice('dropoutRate', [0.1, 0.2, 0.3,0.4]) if debug=='Object' else 0.2        
        args['attntnThreshold'] = 0# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
                
        # Losses
        args['ContrastACC_Pheno'] = 0#4 if debug=='Index' else hp.choice("orthoColor_Lambda0", [1,0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                
        args['ContrastACC_Neigh'] = 0#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['ContrastACC_Area'] = 0#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['PatchEntropy_Pheno'] = 0#3 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda0", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.001   
        args['PatchEntropy_Neigh'] = 0#4 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda1", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0
        args['PatchEntropy_Area'] = 0#1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda2", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.1          
        
        # Bioinsights parameters
        args['Bio_save_Orig_im'] = False

    else:
        
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # or resnet18
        args['PCL_N_Workers']=8 # Number of data loading workers 
        args['PCL_Epochs']=80 # Number of total epoch to run
        args['PCL_N_Crops_per_Image']=40 # Number of crops chosen per image
        args['PCL_Patch_Size']=30 # Size of patches in pixels
        args['PCL_Stride']=0 # Size of stride between consecutive patches
        args['PCL_Alpha_L']=1.15 # size ratio between image crop and image patch
        args['PCL_Z_Score']= True # Whether to apply z score per channel/marker
        args['PCL_Batch_Size']= 1 # Number of images per iteration
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of patch
        args['PCL_Learning_Rate']= 0.003 # Learning Rate
        args['PCL_Temperature']= 0.07 # Softmax temperature    
        args['PCL_eliminate_Black_Background'] = False
        args['PCL_eliminate_White_Background'] = False            
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # GPU where the training is carried out 

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['Clinical_Benefit'] 

        # Architecture Search parameters
        args['num_samples_architecture_search'] = 1500  

        # Optimization Parameters
        args['epochs'] = 1000# if debug else hp.quniform('epochs', 5, 25, 1)
        args['lr_decay_factor'] = 0.1# if debug else hp.uniform('lr_decay_factor', 0, 0.75)
        args['lr_decay_step_size'] = 10000#max(int(args['epochs']/5),1)# if debug else hp.quniform('lr_decay_step_size', 2, 20, 1)        
        args['weight_decay'] = 4 if debug=='Index' else hp.choice('weight_decay',[0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0
        args['batch_size'] = 3 if debug=='Index' else hp.choice('batch_size', [2, 4, 8, 24]) if debug=='Object' else 24
        args['lr'] = 0 if debug=='Index' else hp.choice('lr', [1,0.1,0.01,0.001,0.0001]) if debug=='Object' else 1

        # General
        args['training_MODE'] = 'TrainALL' # TrainALL, NestedCrossValidation, CrossValidation # Whether to use nested cross validation or cross validation.        
        args['folds'] = 1
        args['device'] = 'cuda:0'
        args['Batch_Normalization'] = True
        args['dataAugmentationPerc'] = 0 #0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['N_Phenotypes'] = 8#2 if debug=='Index' else hp.choice('clusters1',[5,10,15,20]) if debug=='Object' else 15     
        args['N_Neighborhoods'] = 16#1 if debug=='Index' else hp.choice('clusters2',[6,11,16,21]) if debug=='Object' else 11 
        args['N_Areas'] = 4#0 if debug=='Index' else hp.choice('clusters3',[2,4,7]) if debug=='Object' else 2      
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = True
        args['Area_Learning'] = True
        args['dropoutRate'] = 0# 1 if debug=='Index' else hp.choice('dropoutRate', [0.1, 0.2, 0.3,0.4]) if debug=='Object' else 0.2        
        args['attntnThreshold'] = 0# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
                
        # Losses
        args['ContrastACC_Pheno'] = 0.01#4 if debug=='Index' else hp.choice("orthoColor_Lambda0", [1,0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                
        args['ContrastACC_Neigh'] = 0.01#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['ContrastACC_Area'] = 0.0001#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['PatchEntropy_Pheno'] = 0.01#3 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda0", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.001   
        args['PatchEntropy_Neigh'] = 0.01#4 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda1", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0
        args['PatchEntropy_Area'] = 0.0001#1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda2", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.1          
        
        # Bioinsights parameters
        args['Bio_save_Orig_im'] = True


    return args


from pycox.models import LogisticHazard
from pycox.evaluation import EvalSurv



# Show example
import numpy as np
import matplotlib.pyplot as plt

# For preprocessing
from sklearn.preprocessing import StandardScaler
from sklearn_pandas import DataFrameMapper 

import torch # For building the networks 
import torchtuples as tt # Some useful functions

from pycox.datasets import metabric
from pycox.evaluation import EvalSurv
from pycox.models import DeepHitSingle

import numpy as np
import pandas as pd
import torch
import torchtuples as tt

df_train = metabric.read_df()
df_test = df_train.sample(frac=0.2)
df_train = df_train.drop(df_test.index)
df_val = df_train.sample(frac=0.2)
df_train = df_train.drop(df_val.index)

# Feature Transforms
cols_standardize = ['x0', 'x1', 'x2', 'x3', 'x8']
cols_leave = ['x4', 'x5', 'x6', 'x7']
standardize = [([col], StandardScaler()) for col in cols_standardize]
leave = [(col, None) for col in cols_leave]
x_mapper = DataFrameMapper(standardize + leave)
x_train = x_mapper.fit_transform(df_train).astype('float32')
x_val = x_mapper.transform(df_val).astype('float32')
x_test = x_mapper.transform(df_test).astype('float32')

# Label Transforms
num_durations = 10
labtrans = DeepHitSingle.label_transform(num_durations)
get_target = lambda df: (df['duration'].values, df['event'].values)
y_train = labtrans.fit_transform(*get_target(df_train))
y_val = labtrans.transform(*get_target(df_val))
train = (x_train, y_train)
val = (x_val, y_val)
type(labtrans)

# NeuralNet
in_features = x_train.shape[1]
num_nodes = [32, 32]
out_features = labtrans.out_features
batch_norm = True
dropout = 0.1
net = tt.practical.MLPVanilla(in_features, num_nodes, out_features, batch_norm, dropout)
model = DeepHitSingle(net, tt.optim.Adam, alpha=0.2, sigma=0.1, duration_index=labtrans.cuts)
batch_size = 256
lr_finder = model.lr_finder(x_train, y_train, batch_size, tolerance=3)
_ = lr_finder.plot()
lr_finder.get_best_lr()
model.optimizer.set_lr(0.01)
epochs = 100
callbacks = [tt.callbacks.EarlyStopping()]
log = model.fit(x_train, y_train, batch_size, epochs, callbacks, val_data=val)

# Prediction
surv = model.predict_surv_df(x_test)
surv = model.interpolate(10).predict_surv_df(x_test)
surv.iloc[:, :5].plot(drawstyle='steps-post')
plt.ylabel('S(t | x)')
_ = plt.xlabel('Time')
ev = EvalSurv(surv, durations_test, events_test, censor_surv='km')
ev.concordance_td('antolini')


class SurvivalAnalysis():
    ''' Code to call pycox Functions and used them'''
    def __init__(self):
        self.num_durations = 10
        self.labtrans = LogisticHazard.label_transform(self.num_durations)
        return

    def labelTransform(self):
        y_train = self.labtrans.fit_transform()


class BaseSimCLRException(Exception):
    """Base exception"""


class InvalidBackboneError(BaseSimCLRException):
    """Raised when the choice of backbone Convnet is invalid."""


class InvalidDatasetSelection(BaseSimCLRException):
    """Raised when the choice of dataset is invalid."""



import torch.nn as nn
import torchvision.models as models

from NaroNet.Patch_Contrastive_Learning.exceptions.exceptions import InvalidBackboneError


class ResNetSimCLR_pretrained(nn.Module):

    def __init__(self, in_channels, base_model, out_dim):
        super(ResNetSimCLR_pretrained, self).__init__()
        self.resnet_dict = {"resnet50": models.resnet50(pretrained=True)}#, num_classes=out_dim)}

        self.backbone = self._get_basemodel(base_model)        

        self.backbone.conv1 = nn.Conv2d(in_channels, self.backbone.conv1.out_channels, self.backbone.conv1.kernel_size, 
                                        stride=self.backbone.conv1.stride, padding=self.backbone.conv1.padding)

        # add mlp projection head
        #self.backbone.fc = nn.Sequential(nn.Linear(out_dim, 1028), nn.ReLU(), nn.Linear(1000, out_dim), nn.ReLU(), self.backbone.fc)
        self.backbone.fc = nn.Sequential(self.backbone.fc,nn.ReLU(),nn.Linear(1000, out_dim),nn.ReLU() ,nn.Linear(out_dim, 1028))

    def _get_basemodel(self, model_name):
        try:
            model = self.resnet_dict[model_name]
        except KeyError:
            raise InvalidBackboneError(
                "Invalid backbone architecture. Check the config file and pass one of: resnet18 or resnet50")
        else:
            return model

    def forward(self, x):
        return self.backbone(x)



class ResNetSimCLR(nn.Module):

    def __init__(self, in_channels, base_model, out_dim):
        super(ResNetSimCLR, self).__init__()
        self.resnet_dict = {"resnet18": models.resnet18(pretrained=False, num_classes=out_dim),
                            "resnet50": models.resnet50(pretrained=False, num_classes=out_dim)}

        self.backbone = self._get_basemodel(base_model)
        dim_mlp = self.backbone.fc.in_features

        self.backbone.conv1 = nn.Conv2d(in_channels, self.backbone.conv1.out_channels, self.backbone.conv1.kernel_size, 
                                        stride=self.backbone.conv1.stride, padding=self.backbone.conv1.padding)

        # add mlp projection head
        self.backbone.fc = nn.Sequential(nn.Linear(dim_mlp, dim_mlp), nn.ReLU(), self.backbone.fc)

    def _get_basemodel(self, model_name):
        try:
            model = self.resnet_dict[model_name]
        except KeyError:
            raise InvalidBackboneError(
                "Invalid backbone architecture. Check the config file and pass one of: resnet18 or resnet50")
        else:
            return model

    def forward(self, x):
        return self.backbone(x)




import os
from ray import tune
from ray.tune import Trainable
from ray.tune.schedulers import ASHAScheduler
from ray.tune.suggest.hyperopt import HyperOptSearch
from ray.tune.suggest import ConcurrencyLimiter
from NaroNet.architecture_search.extract_info_architectureSearch import extract_best_result
from NaroNet.architecture_search.extract_info_architectureSearch import save_architecture_search_stats
import torch
import ray
from NaroNet import NaroNet
from NaroNet.NaroNet_simple import NaroNet_simple

class NaroNet_Search(Trainable):


    def _setup(self, parameters):  
        self.device = torch.device(parameters["device"] if torch.cuda.is_available() else "cpu")
        # Initialize 
        self.N = NaroNet.NaroNet(parameters, self.device)        
        self.N.epoch = 0
              
    def _train(self):
        try:
            result = self.N.epoch_validation()
        except Exception as exc:
            del self.N                                   
            result = {"val_acc":0.1,"epoch": 1000,"maximize_acc_interpretability":0,"acc_test":0,"interpretability":0,"train_Cell_ent":0,"train_Cross_entropy":10,"train_Pat_ent":0,"train_UnsupContrast_acc":0,"train_acc":0,"train_loss":0,"val_Cell_ent":0,"val_Pat_ent":0,"val_UnsupContrast_acc":0,"val_loss":10,'val_Cross_entropy':10,'test_Cross_entropy':10}             
            pass            
        return result

    def _save(self, checkpoint_dir):        
        return

class NaroNet_Search_simple(Trainable):
    def _setup(self, parameters):  
        self.device = torch.device(parameters["device"] if torch.cuda.is_available() else "cpu")
        # Initialize 
        self.N = NaroNet_simple(parameters, self.device)
        self.N.epoch = 0
              
    def _train(self):
        try:
            result = self.N.epoch_validation()
        except Exception as exc:
            del self.N                                   
            result = {"val_acc":0.1,"epoch": 1000,"cross_ent_test":1000,"acc_test":0,"interpretability":0,"train_Cell_ent":0,"train_Cross_entropy":10,"train_Pat_ent":0,"train_UnsupContrast_acc":0,"train_acc":0,"train_loss":0,"val_Cell_ent":0,"val_Pat_ent":0,"val_UnsupContrast_acc":0,"val_loss":10,'val_Cross_entropy':10,'test_Cross_entropy':10}             
            pass            
        return result

    def _save(self, checkpoint_dir):        
        return


def architecture_search(path,best_parameters,possible_parameters,AvoidSet=None):
    
    # In case NestedCrossvalidation, add AvoidSet to the parameters.
    if AvoidSet is not None:
        possible_parameters['AvoidSet'] = AvoidSet

    # Metric to optimize
    metric = 'cross_ent_test' #  "loss_test"
    num_gpus = 1
    best_parameters['device'] = best_parameters['device'] if 'Synthetic' in path else 'cuda'
    possible_parameters['device'] = possible_parameters['device'] if 'Synthetic' in path else 'cuda'

    # Whether to search architecture or load previous ones.
    architecture_search_path = path+'Architecture_Search/'
    architecture_search_path_save = architecture_search_path+'Results/'
    if (not os.path.exists(architecture_search_path)) or (not os.path.exists(architecture_search_path+'NaroNet_Search/')) or (not os.path.exists(architecture_search_path_save)):        
        os.mkdir(architecture_search_path)
        os.mkdir(architecture_search_path+'NaroNet_Search/')
        os.mkdir(architecture_search_path_save)
    else:
        best_result, n_runs = extract_best_result(architecture_search_path+'NaroNet_Search/',metric,best_parameters)
        if (n_runs>10 and len(os.listdir(architecture_search_path_save))==0):
            save_architecture_search_stats(architecture_search_path_save,architecture_search_path+'NaroNet_Search/',10)
        if n_runs>best_parameters['num_samples_architecture_search']*0.9:
            return best_result

    # Restart Ray defensively in case the ray connection is lost. 
    # ray start --head --node-ip-address=127.0.0.1 --port=34009
    # ray start --address='128.3.60.116:34009' --redis-password='5241590000000000'
    # ray.shutdown()  
    ray.init(local_mode=True)#,_redis_password='sadfh*&H^&T^&TU')#address='127.0.0.1:34009', _redis_password='5241590000000000')#) # Local address #,address='127.0.0.1:6379'
    
    # Set Scheduler
    scheduler=ASHAScheduler(time_attr="epoch",max_t=best_parameters['epochs']-2, metric=metric, mode="min")        
    
    # Set algorithm
    search_algo = HyperOptSearch(possible_parameters, metric=metric, mode="min",n_initial_points=int(best_parameters['num_samples_architecture_search']*0.8), points_to_evaluate=[best_parameters])    
    search_algo = ConcurrencyLimiter(search_algo, max_concurrent=1)

    # Running parameters
    runIt = {"num_samples": best_parameters['num_samples_architecture_search'], 
    "resources_per_trial":{"gpu": 0, "cpu":1}, "checkpoint_freq":0, "local_dir":architecture_search_path} 
    sync_config = tune.SyncConfig()
        
    # Obtain the best hyperparametersWWW
    analysis = tune.run(NaroNet_Search,scheduler=scheduler,search_alg=search_algo,sync_config=sync_config,**runIt)

    # Obtain best result parameters
    best_result, n_runs = extract_best_result(architecture_search_path+'NaroNet_Search/',metric,best_parameters)
    return best_result



from tkinter import N
from torchvision.transforms import transforms
from NaroNet.Patch_Contrastive_Learning.data_aug.gaussian_blur import GaussianBlur
from torchvision import transforms, datasets
from NaroNet.Patch_Contrastive_Learning.data_aug.view_generator import ContrastiveLearningViewGenerator
from NaroNet.Patch_Contrastive_Learning.exceptions.exceptions import InvalidDatasetSelection
from NaroNet.utils import utilz
from imgaug import augmenters as iaa
from torch.utils.data import Dataset, DataLoader 
import pandas as pd
import torch
import os
import numpy as np
from skimage import io
from skimage import filters
import random
import gc
import csv
from NaroNet.Patch_Contrastive_Learning.preprocess_images import Mean_std_experiment

class ContrastiveLearningDataset_PCL(Dataset):
    def __init__(self, args,training, mean=None, std=None):
        self.training = training

        # Root folder
        self.root_folder = args['path']+'Raw_Data/Images/'

        # Image list
        self.image_list = os.listdir(self.root_folder)
        random.shuffle(self.image_list)

        # Open Channels.txt
        self.Channels, _ = utilz.load_channels(args['path']+'Raw_Data/')
        self.in_channels = len(self.Channels)        

        # Parameters
        self.args = args    
        self.patch_size = args['PCL_Patch_Size']
        self.stride = args['PCL_Stride']
        self.alpha_L = args['PCL_Alpha_L']
        self.n_crops_per_image = args['PCL_N_Crops_per_Image']    

        # Preprocessing
        if os.path.exists(self.root_folder[:-7]+'Experiment_Stats.csv'):
            f = csv.reader(open(self.root_folder[:-7]+'Experiment_Stats.csv'))
            self.mean = np.array(next(f),dtype=np.float32)
            self.std = np.array(next(f),dtype=np.float32)            
        else: 
            self.mean, self.std = Mean_std_experiment(obj=self, base_path=self.root_folder,image_paths=self.image_list[:10],Channels=self.Channels)     
    
    def __len__(self):        
        return len(self.image_list)

    def load_image(self,idx):

        # Load CZI file
        if self.image_list[idx].endswith('.czi'):
            from aicspylibczi import CziFile
            import pathlib
            import cv2
            
            # Obtain header info from the czi
            factor = 5
            file = pathlib.Path(self.root_folder+self.image_list[idx])
            czi = CziFile(file)
            self.czidimensions = czi.get_mosaic_scene_bounding_box(0)    
            dims = czi.get_dims_shape()        
            
            # Obtain crop or the full image
            if self.training:
                xIn,yIn = np.random.randint(0,self.czidimensions.w-int(self.patch_size*factor*self.alpha_L)-1), np.random.randint(0,self.czidimensions.h-int(self.patch_size*factor*self.alpha_L)-1)
                image = czi.read_mosaic(C=0,region=(self.czidimensions.x+xIn,self.czidimensions.y+yIn,int(self.patch_size*factor*self.alpha_L),int(self.patch_size*factor*self.alpha_L)))
                new_size_x, new_size_y  = int(self.patch_size*self.alpha_L), int(self.patch_size*self.alpha_L)
            else:    
                image = czi.read_mosaic(C=0,region=(self.czidimensions.x,self.czidimensions.y,self.czidimensions.w,self.czidimensions.h))
                new_size_x, new_size_y  = int(self.czidimensions.w/factor), int(self.czidimensions.h/factor)
            
            # Postprocessing and image resizing
            image = image.squeeze()
            image = cv2.resize(image, dsize=(new_size_x, new_size_y), interpolation=cv2.INTER_CUBIC)
            self.wsiIm = czi

        elif self.image_list[idx].endswith('.ome.tif'): 
            image = io.imread(self.root_folder+self.image_list[idx])              
            image = image[:,:,self.Channels]
            if self.args['PCL_Z_Score'] and hasattr(self, 'mean'):
                image -= np.expand_dims(np.expand_dims(self.mean,0),0)
                image /= np.expand_dims(np.expand_dims(self.std,0),0)

        # Load TIFF file
        elif self.image_list[idx].endswith('.tif') or self.image_list[idx].endswith('.tiff'): 
            image = io.imread(self.root_folder+self.image_list[idx])                
            
            if np.argmin(image.shape)==0:
                image = image.swapaxes(0,2)                            
            
            image = image[:,:,self.Channels]
            if self.args['PCL_Z_Score'] and hasattr(self, 'mean'):
                image = image-np.expand_dims(np.expand_dims(self.mean,0),0)
                image = image/np.expand_dims(np.expand_dims(self.std,0),0)
        else:
            image =  np.load(self.root_folder+self.image_list[idx],allow_pickle=True)  
            from PIL import Image            
        return image

    def get_crops(self, data,idx):

        # Don´t train using 
        thr = filters.threshold_multiotsu(data.sum((2)),classes=3)
        
        images=[]
        for _ in range(int(self.n_crops_per_image)):
            
            if self.image_list[idx].endswith('.czi'):                
                images.append(self.load_image(idx))
            else:                                
                while True:
                
                    # Get random x,y positions within the image and append the local phenotype crop
                    xIn,yIn = np.random.randint(0,data.shape[0]-int(self.patch_size*self.alpha_L)-1), np.random.randint(0,data.shape[1]-int(self.patch_size*self.alpha_L)-1)
                    
                    if data[xIn:xIn+int(self.patch_size*self.alpha_L),yIn:yIn+int(self.patch_size*self.alpha_L),:].sum((2)).max()>=thr[0] and self.args['PCL_eliminate_Black_Background']:                                        
                        images.append(data[xIn:xIn+int(self.patch_size*self.alpha_L),yIn:yIn+int(self.patch_size*self.alpha_L),:])
                        break
                    elif data[xIn:xIn+int(self.patch_size*self.alpha_L),yIn:yIn+int(self.patch_size*self.alpha_L),:].sum((2)).max()<=thr[0] and self.args['PCL_eliminate_White_Background']:                                        
                        images.append(data[xIn:xIn+int(self.patch_size*self.alpha_L),yIn:yIn+int(self.patch_size*self.alpha_L),:])
                        break
                    elif self.args['PCL_eliminate_Black_Background']==False and self.args['PCL_eliminate_Black_Background']==False: 
                        images.append(data[xIn:xIn+int(self.patch_size*self.alpha_L),yIn:yIn+int(self.patch_size*self.alpha_L),:])
                        break        
        return images  

    def get_all_crops(self, data):
        
        # Obtain indices considering patch size and stride
        patches=[]  
        patch_pos = []     
        x_steps = range(int(np.floor(self.patch_size/2)), int(np.floor(data.shape[0]/(self.patch_size-self.stride))*(self.patch_size-self.stride)), (self.patch_size-self.stride))
        y_steps = range(int(np.floor(self.patch_size/2)), int(np.floor(data.shape[1]/(self.patch_size-self.stride))*(self.patch_size-self.stride)), (self.patch_size-self.stride))      

        # Iterate to obtain patches
        for x in x_steps:
            for y in y_steps:
                patches.append(data[int(np.ceil(x-self.patch_size/2)):int(np.ceil(x+self.patch_size/2)),
                                    int(np.ceil(y-self.patch_size/2)):int(np.ceil(y+self.patch_size/2)),:])            
                patch_pos.append([x,y])
        
        return patches, np.array(patch_pos)

    def perform_augmentation(self, crop):

        # Obtain image patch from image crop.
        x, y =np.random.randint(0,crop[0].shape[0]-self.patch_size+1), np.random.randint(0,crop[0].shape[0]-self.patch_size+1)        
        aug = crop[x:self.patch_size+x,y:self.patch_size+y,:]

        # Add color perturbation
        if self.args['PCL_Color_perturbation_Augmentation']:
            # To install the following library: pip install histomicstk --find-links https://girder.github.io/large_image_wheels
            from histomicstk.preprocessing.augmentation.color_augmentation import rgb_perturb_stain_concentration
            min_im = aug.min((0,1))
            aug = aug-min_im
            max_im = aug.max((0,1))
            aug = aug/max_im*255        
            aug = rgb_perturb_stain_concentration(aug)
            aug = (((aug*max_im)/255)+min_im)
        
        # Define sequence of augmentations
        seq_easy = iaa.Sequential([
            iaa.CoarseDropout(0.04, size_percent=0.05, per_channel=0.5), # Drop 2% of all pixels, 15% of the original size,in 50% of all patches channeld
            iaa.Add((-0.15, 0.15), per_channel=0.5), # Add noise per channel...
            iaa.GaussianBlur(sigma=(0, 1)) # blur patches with a sigma of 0 to 3.0
        ])                
        
        # Flip lr
        if random.random()<0.5:
            for c in range(aug.shape[2]):
                aug[:,:,c] = np.flip(aug[:,:,c],0)
    
        # Flip ud
        if random.random()<0.5:
            for c in range(aug.shape[2]):
                aug[:,:,c] = np.flip(aug[:,:,c],1)

        # Color_Distortion...->...Sobel...->...Gaussian_Noise...->...Rotate        
        return seq_easy(images=aug.astype(np.float32))    

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        # Load image
        image = self.load_image(idx)    
        # image = image.astype(np.float32)    

        if self.training:

            # Obtain a crop of size Patch_size*alpha_L
            crops = self.get_crops(image,idx)
            
            # Perform sequential image transformations
            augs_1 = []
            augs_2 = []
            for crop in crops:
                augs_1.append(self.perform_augmentation(crop))
                augs_2.append(self.perform_augmentation(crop))
            del image, crops
            gc.collect(0)
            gc.collect(1)
            gc.collect(2)
            return torch.tensor(np.array(augs_1)), torch.tensor(np.array(augs_2))

        else:
            crops, patch_pos = self.get_all_crops(image)
            del image
            gc.collect(0)
            gc.collect(1)
            gc.collect(2)
            return crops, patch_pos, self.image_list[idx]


        





import numpy as np
import copy
import seaborn as sns
import os.path as osp
import matplotlib.pyplot as plt
import pandas as pd

def save_celltype_abundance_per_patient(dataset, IndexAndClass, clusters, heatmapClusterPresence, labels, heatmap_Name,ticklabel):    

    for idx, ClusterLevel in enumerate(clusters):        
        
        list_of_labels = [i[0] for i in labels]

        # Save heatmap as excel file.
        dict_to_Save = {'Patient_Names':[ia[0] for ia in IndexAndClass],'Patient_Labels':list_of_labels}
        for TME in range(heatmapClusterPresence[str(ClusterLevel)].shape[1]):
            dict_to_Save[ticklabel[idx]+str(TME+1)]=list(heatmapClusterPresence[str(ClusterLevel)][:,TME])
        df1 = pd.DataFrame.from_dict(dict_to_Save)
        df1.to_excel(dataset.bioInsights_dir_cell_types_abundance+'heatmap_{}_Abundance.xlsx'.format(heatmap_Name[idx]))  

        # Save color information for next Cluster Level          
        labels_pal = sns.cubehelix_palette(len(set(list_of_labels)), light=.9, dark=.1, reverse=True, start=0, rot=-2)
        labels_lut = dict(zip(map(str, list(set(list_of_labels))), labels_pal))        
        labels_colors = [labels_lut[str(k)] for k in list_of_labels]
        
        # Create ClusterMap        
        heatmapClusterPresence_Fig = sns.clustermap(heatmapClusterPresence[str(ClusterLevel)],col_cluster=True, 
                                        row_colors=labels_colors, linewidths=0, cmap="vlag",
                                        xticklabels=[ticklabel[idx]+str(ind+1) for ind in range(ClusterLevel)])            
            
        # Save heatmap
        for label in list(set(list_of_labels)):
            heatmapClusterPresence_Fig.ax_col_dendrogram.bar(0, 0, color=labels_lut[str(label)], label=label, linewidth=0)
        heatmapClusterPresence_Fig.ax_col_dendrogram.legend(title=dataset.experiment_label[0])#, loc="left")#, ncol=5, bbox_to_anchor=(0.47, 0.8), bbox_transform=plt.gcf().transFigure)
        heatmapClusterPresence_Fig.savefig(dataset.bioInsights_dir_cell_types_abundance+'heatmap_{}_Abundance.png'.format(heatmap_Name[idx]),dpi=1000) 

def obtain_celltype_abundance_per_patient(dataset, IndexAndClass, clusters, ClusterThreshold):
    # Initialize heatmap Cluster Presence 
    if len(clusters)==1:
        heatmapClusterPresence = {str(clusters[0]):[]}        
    elif len(clusters)==2:
        heatmapClusterPresence = {str(clusters[0]):[], str(clusters[1]):[]}        
    elif len(clusters)==3:
        heatmapClusterPresence = {str(clusters[0]):[], str(clusters[1]):[], str(clusters[2]):[]}        
    labels = copy.deepcopy(heatmapClusterPresence)
    for idx, ClusterLevel in enumerate(clusters):
        heatmapClusterPresence[str(ClusterLevel)] = np.zeros([len(IndexAndClass),ClusterLevel])        
    
    # Obtain clusters per Slide
    for count, idxclster in enumerate(IndexAndClass):       
        # Obtain clusters per cluster Level
        for idx, ClusterLevel in enumerate(clusters):
            # Load Cluster Assignment for a specific image
            if idx<2:
                cluster_assignment_raw = np.load(dataset.processed_dir_cell_types+'cluster_assignmentPerPatch_Index_{}_{}_ClustLvl_{}.npy'.format(idxclster[1],0, ClusterLevel))
                cluster_assignment_raw[np.percentile(cluster_assignment_raw.max(-1),ClusterThreshold)>cluster_assignment_raw]=1e-16
                cluster_assignment_raw = cluster_assignment_raw.sum(-2)                
            else:
                cluster_assignment_raw = np.load(dataset.processed_dir_cell_types+'cluster_assignment_Index_{}_ClustLvl_{}.npy'.format(idxclster[1], ClusterLevel))                                            
            if idx>0:
                cluster_assignment_prev = copy.deepcopy(cluster_assignment)
                cluster_assignment = copy.deepcopy(cluster_assignment_raw)
            else:
                cluster_assignment = copy.deepcopy(cluster_assignment_raw)
            
            # Save value to heatmap
            if idx==0:
                heatmapClusterPresence[str(ClusterLevel)][count,:] = cluster_assignment_raw/(cluster_assignment_raw.sum()+1e-16)
            else:
                if len(cluster_assignment_raw.shape)==1:
                    heatmapClusterPresence[str(ClusterLevel)][count,:] = cluster_assignment_raw/(cluster_assignment_raw.sum()+1e-16)
                else:
                    clustersNOW=cluster_assignment_prev # Eliminate clusters with zeros attention from the previous iteration.
                    cluster_assignment = np.matmul(clustersNOW,cluster_assignment)
                    heatmapClusterPresence[str(ClusterLevel)][count,:] = (cluster_assignment/(cluster_assignment.sum()+1e-16))                        
            # Assign clusters to values            
            labels[str(ClusterLevel)].append(idxclster[2])
    return heatmapClusterPresence, labels



import logging
import os
import sys

import torch
import torch.nn.functional as F
from torch.cuda.amp import GradScaler, autocast
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm
from utils import save_config_file, accuracy, save_checkpoint

torch.manual_seed(0)


class SimCLR(object):

    def __init__(self, *args, **kwargs):
        self.args = kwargs['args']
        self.model = kwargs['model'].to(self.args.device)
        self.optimizer = kwargs['optimizer']
        self.scheduler = kwargs['scheduler']
        self.writer = SummaryWriter()
        logging.basicConfig(filename=os.path.join(self.writer.log_dir, 'training.log'), level=logging.DEBUG)
        self.criterion = torch.nn.CrossEntropyLoss().to(self.args.device)

    def info_nce_loss(self, features):

        labels = torch.cat([torch.arange(self.args.batch_size) for i in range(self.args.n_views)], dim=0)
        labels = (labels.unsqueeze(0) == labels.unsqueeze(1)).float()
        labels = labels.to(self.args.device)

        features = F.normalize(features, dim=1)

        similarity_matrix = torch.matmul(features, features.T)
        # assert similarity_matrix.shape == (
        #     self.args.n_views * self.args.batch_size, self.args.n_views * self.args.batch_size)
        # assert similarity_matrix.shape == labels.shape

        # discard the main diagonal from both: labels and similarities matrix
        mask = torch.eye(labels.shape[0], dtype=torch.bool).to(self.args.device)
        labels = labels[~mask].view(labels.shape[0], -1)
        similarity_matrix = similarity_matrix[~mask].view(similarity_matrix.shape[0], -1)
        # assert similarity_matrix.shape == labels.shape

        # select and combine multiple positives
        positives = similarity_matrix[labels.bool()].view(labels.shape[0], -1)

        # select only the negatives the negatives
        negatives = similarity_matrix[~labels.bool()].view(similarity_matrix.shape[0], -1)

        logits = torch.cat([positives, negatives], dim=1)
        labels = torch.zeros(logits.shape[0], dtype=torch.long).to(self.args.device)

        logits = logits / self.args.temperature
        return logits, labels

    def train(self, train_loader):

        scaler = GradScaler(enabled=self.args.fp16_precision)

        # save config file
        save_config_file(self.writer.log_dir, self.args)

        n_iter = 0
        logging.info(f"Start SimCLR training for {self.args.epochs} epochs.")
        logging.info(f"Training with gpu: {self.args.disable_cuda}.")

        for epoch_counter in range(self.args.epochs):
            for images, _ in tqdm(train_loader):
                images = torch.cat(images, dim=0)

                images = images.to(self.args.device)

                with autocast(enabled=self.args.fp16_precision):
                    features = self.model(images)
                    logits, labels = self.info_nce_loss(features)
                    loss = self.criterion(logits, labels)

                self.optimizer.zero_grad()

                scaler.scale(loss).backward()

                scaler.step(self.optimizer)
                scaler.update()

                if n_iter % self.args.log_every_n_steps == 0:
                    top1, top5 = accuracy(logits, labels, topk=(1, 5))
                    self.writer.add_scalar('loss', loss, global_step=n_iter)
                    self.writer.add_scalar('acc/top1', top1[0], global_step=n_iter)
                    self.writer.add_scalar('acc/top5', top5[0], global_step=n_iter)
                    self.writer.add_scalar('learning_rate', self.scheduler.get_lr()[0], global_step=n_iter)

                n_iter += 1

            # warmup for the first 10 epochs
            if epoch_counter >= 10:
                self.scheduler.step()
            logging.debug(f"Epoch: {epoch_counter}\tLoss: {loss}\tTop1 accuracy: {top1[0]}")

        logging.info("Training has finished.")
        # save model checkpoints
        checkpoint_name = 'checkpoint_{:04d}.pth.tar'.format(self.args.epochs)
        save_checkpoint({
            'epoch': self.args.epochs,
            'arch': self.args.arch,
            'state_dict': self.model.state_dict(),
            'optimizer': self.optimizer.state_dict(),
        }, is_best=False, filename=os.path.join(self.writer.log_dir, checkpoint_name))
        logging.info(f"Model checkpoint and metadata has been saved at {self.writer.log_dir}.")



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="NaroNet", 
    version="1.0.17",
    author="Daniel Jiménez-Sánchez",
    author_email="danijimnzs@gmail.com",
    description="NaroNet: discovery of tumor microenvironment elements from multiplex images.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/djimenezsanchez/NaroNet",
    project_urls={
        "Bug Tracker": "https://github.com/djimenezsanchez/NaroNet/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
		"License :: OSI Approved :: BSD License",
    ],
	license='BSD 3-Clause License',
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.7",
    install_requires = [                                     
        'aicspylibczi>=3.0.5','czifile>=2019.7.2','imageio>=2.16.0',
        'hyperopt>=0.2.3','imgaug','matplotlib>=3.2.1',
        'numpy>=1.21.2','opencv-python>=4.5.5','pandas','imblearn',
        'torch','tensorboard>=2.8.0','scikit-learn','scikit-image',
        'Pillow','tifffile>=2020.2.9','xlsxwriter>=3.0.2','tqdm>=4.50.2','xlrd>=1.2.0',
        'argparse>=1.1','seaborn>=0.11.0','scipy>=1.5.4','openTSNE'
    ]
)


# Upload your package to PyPi
# Now, we create a source distribution with the following command:
# python setup.py sdist

# To upload to pypi test
# twine upload --repository testpypi dist/*
# To upload to pypi
# twine upload dist/*
# You will be asked to provide your username and password. Provide the credentials you used to register to PyPi earlier.